#ifndef lint
static char rcsid[] =
   "$Xukc: execute.c,v 1.7 91/06/23 11:26:05 rlh2 Exp $";
#endif /* !lint */

/* 
 * Copyright 1990 Richard Hesketh / rlh2@ukc.ac.uk
 *                Computing Lab. University of Kent at Canterbury, UK
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of Richard Hesketh and The University of
 * Kent at Canterbury not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Richard Hesketh and The University of Kent at Canterbury make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * Richard Hesketh AND THE UNIVERSITY OF KENT AT CANTERBURY DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL Richard Hesketh OR THE
 * UNIVERSITY OF KENT AT CANTERBURY BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 *
 * Author:  Richard Hesketh / rlh2@ukc.ac.uk, 
 *                Computing Lab. University of Kent at Canterbury, UK
 */

/* Contains all the routines for forking and executing the perly script.
 * Two pipes are opened up between the executing script and the parent and
 * the standard output and standard errors generated by the script are sent down
 * these pipes.
 */

#include <X11/Xlib.h>
#include <X11/Xos.h>
#include <stdio.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <X11/Intrinsic.h>
#include <X11/Xukc/memory.h>
#include "perly.h"

#ifndef DEFAULT_SHELL
#	define DEFAULT_SHELL "#!/bin/sh\n"
#endif /* !DEFAULT_SHELL */

extern void FinishedScript();

int	NeedChdir = 0;		/* whether we need a chdir */
char	cwd[1024];		/* Current directory */

struct	passwd	*myPwd;		/* Owners password entry */


static int	invocations = 0;
static char	stem[] = "/tmp/perly%dRun%d";
static char	scriptstem[50];

extern char	*getenv(/* char* */);
extern char	*malloc(/* int */);
extern char	*index();
extern void	SetDirectory();
extern int	getpid();

/*  return sys_errlist[errno] if in range
 */
char *
reason()
{
	extern char *sys_errlist[];
	extern int errno, sys_nerr;

	return((errno > 0 && errno < sys_nerr) ? sys_errlist[errno]
							: "unknown reason");
}


/* D.I.Y version of mktemp
 * ||| we could use the host id from "gethostid()" as well?? |||
 */
char *
my_mktemp(number)
int number;
{
	extern char *sprintf();

	return (sprintf(scriptstem, stem, getpid(), number));
}


/*
 *	Do a cd action
 *	Return 1 on failure
 *		0 on success
 */
static int
do_chdir(towhere)
char	*towhere;
{
	register struct passwd *pwd;
	register char *s;
	register char c;

	if (NeedChdir) {
		NeedChdir = 0;
		if (*towhere == '~') {
			pwd = (struct passwd *)NULL;
			s = index(towhere, '/');
			c = 0;
			if (s) {
				c = *s;
				*s = '\0';
			}

			if (myPwd == NULL)
				myPwd = getpwuid(getuid());

			if (!s || s == towhere+1)
				/* just the users home directory "~" */
				pwd = myPwd;
			else if (myPwd &&
				 (strcmp(towhere+1, myPwd->pw_name) == 0))
					pwd = myPwd;

			if (pwd == (struct passwd *)NULL)
				pwd = getpwnam(towhere+1);
			if (s)
				*s++ = c;
			if (pwd == (struct passwd *)NULL)
				return(1);
			if (chdir(pwd->pw_dir) < 0)
				return(1);
			if (c && chdir(s) < 0)
				return(1);
		} else if (chdir(towhere) < 0)
				return(1);
	}
	return(0);
}


/*
 *	Pick up children
 */
static void
child_done()
{
	union wait status;
	register int pid;
	
	while ((pid  = wait3(&status, WNOHANG, (struct rusage *)NULL)) > 0);
}


/************* Module Exported Routines **************/

/*
 *	Do the interrupt action
 */
void
SendSignal(butt, signal)
struct app_resources *butt;
int signal;
{
	if (butt->pid == 0)
		return;
	killpg(butt->pid, signal);
}


/*
 *      Run the Perly Script
 */
int
RunScript(butt, dpy, cmd, args, num_args)
struct app_resources *butt;
Display *dpy;
char *cmd;
char **args;
int num_args;
{
	int	pid;
	int	stat;
	int	fd;
	fd_set	readfds, fds;
	char	**pass_args;
	int	i;
	int	errpfd[2];
	int	outpfd[2];

	butt->pipefds[0] = 0;
	butt->pipefds[1] = 0;

	if (cmd == NULL || *cmd == '\0')
		return (0);
		
	SetDirectory(butt, cwd);
	NeedChdir = 1;

	FD_ZERO(&fds);
	FD_SET(ConnectionNumber(dpy), &fds);

	butt->scriptfile = my_mktemp(invocations);
	butt->scriptfile = XtNewString(butt->scriptfile);
	butt->pid = NULL;
	invocations++;

	if (pipe(errpfd) == -1) {
		fprintf(stderr, "Cannot open a stderr pipe (%s)\n", reason());
		return (0);
	}

	if (pipe(outpfd) == -1) {
		fprintf(stderr, "Cannot open a stdout pipe (%s)\n", reason());
		close(errpfd[0]);
		close(errpfd[1]);
		return (0);
	}

	if ((fd = open(butt->scriptfile, O_WRONLY|O_APPEND|O_CREAT, 0700)) < 0){
		fprintf(stderr, "Cannot create script file: %s (s)\n",
				butt->scriptfile, reason());
		close(errpfd[0]);
		close(errpfd[1]);
		close(outpfd[0]);
		close(outpfd[1]);
		return (0);
	}

	if (cmd[1] != '!')
		write(fd, DEFAULT_SHELL, strlen(DEFAULT_SHELL));
	write(fd, cmd, strlen(cmd));
	close(fd);

	pass_args = (char **)malloc(sizeof(char *) *
					(num_args + 2));
	if (pass_args == NULL) { /* malloc failed */
		fprintf(stderr, "malloc failed before script was executed\n");
		close(errpfd[0]);
		close(errpfd[1]);
		close(outpfd[0]);
		close(outpfd[1]);
		return (0);
	}

	pass_args[0] = butt->scriptfile;
	for (i = 0; i < num_args; i++)
		pass_args[i+1] = args[i];
	pass_args[num_args+1] = NULL; /* terminate the args */
		
	(void)signal(SIGCHLD, child_done);

	switch (pid = fork()) {
		case -1:
			fprintf(stderr, "Cannot fork\n");
			close(errpfd[0]);
			close(errpfd[1]);
			close(outpfd[0]);
			close(outpfd[1]);
			return 0;
		case 0:
			/*
			 *	Child
			 */
			setpgrp(0, getpid());
			
			close(outpfd[0]);
			close(errpfd[0]);
			if (!butt->use_stdin)
				close(0); /* close stdin */
			if (dup2(outpfd[1], 1) < 0)
				abort();
			if (dup2(errpfd[1], 2) < 0)
				abort();

			if (do_chdir(cwd)) {
				write(2, "Chdir to ", 9);
				write(2, cwd, strlen(cwd));
				write(2, " failed\n", 8);
				exit(1);
			}
				
#if 0
			if ((fd = open(butt->scriptfile, O_RDONLY, 0700)) < 0) {
				write(2, "Cannot read script from ", 24);
				write(2, butt->scriptfile,
					strlen(butt->scriptfile));
				write(2, "\n", 1);
				exit(2);
			} else
				(void)unlink(butt->scriptfile);
#endif

			close(ConnectionNumber(dpy));
			execv(butt->scriptfile, pass_args);
			/* should never get here */
			close(outpfd[1]);
			close(errpfd[1]);
			close(fd);
			exit(1);
		default:
			butt->pid = pid;
			break;
	}
	butt->pipefds[0] = outpfd[0];
	butt->pipefds[1] = errpfd[0];
	close(outpfd[1]);
	close(errpfd[1]);
	free((char *)pass_args);
	return (butt->pid);
}



void
SetDirectory(butt, str)
struct app_resources *butt;
register char *str;
{
	if (butt->execution_directory == NULL)
		butt->execution_directory = XtNewString("~");
	(void)strcpy(str, butt->execution_directory);
}


/*
 *	Pick up children
 */
void
CleanUpOnChildEOF(butt)
struct app_resources *butt;
{
	if (butt->scriptfile != NULL) {
		(void)unlink(butt->scriptfile);
		XtFree(butt->scriptfile);
		butt->scriptfile = NULL;
	}
	SendSignal(butt, SIGHUP);
	butt->pid = NULL;
}
