#ifndef lint
static char rcsid[] =
   "$Xukc: file.c,v 1.9 91/04/15 18:57:25 rlh2 Exp $";
#endif /* !lint */

/* 
 * Copyright 1990 Richard Hesketh / rlh2@ukc.ac.uk
 *                Computing Lab. University of Kent at Canterbury, UK
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of Richard Hesketh and The University of
 * Kent at Canterbury not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Richard Hesketh and The University of Kent at Canterbury make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * Richard Hesketh AND THE UNIVERSITY OF KENT AT CANTERBURY DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL Richard Hesketh OR THE
 * UNIVERSITY OF KENT AT CANTERBURY BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 *
 * Author:  Richard Hesketh / rlh2@ukc.ac.uk, 
 *                Computing Lab. University of Kent at Canterbury, UK
 */


#include <ctype.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <stdio.h>
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/Xukc/Command.h>
#include <X11/Xukc/memory.h>
#include <X11/Xukc/text.h>
#include "perly.h"

extern char *reason();
extern char *getenv();
extern int errno;

extern Window XukcFindUndecoratedWindow();

Boolean StoreAttribute();
void StoreGeometry();

/* wm_offset is a kludge for a broken version of MWM (1.1) */
static Dimension wm_offset = 0;

static void
set_wm_offset(dpy)
Display *dpy;
{
	static Boolean initialized = FALSE;

	if (initialized)
		return;

	initialized = TRUE;

	/* see if the Motif window manager is running */
	if (XInternAtom(dpy, "_MOTIF_WM_INFO", TRUE) != None)
		wm_offset = 2;
}


/* this routines a string that is unique to the server and display */
static String
get_server_id(butt)
struct app_resources *butt;
{
	Display *dpy;
	Screen *screen;
	static char server_str[500];

	dpy = XtDisplay(butt->perly_button);
	screen = XtScreen(butt->perly_button);

	(void)sprintf(server_str, "%s:%d (%d,%d,%dx%d)", ServerVendor(dpy),
			VendorRelease(dpy), ScreenCount(dpy),
			DefaultDepthOfScreen(screen),
			WidthOfScreen(screen), HeightOfScreen(screen));
	return (server_str);
}


static String
cvt_to_filename(butt, full_path, name)
struct app_resources *butt;
String full_path, name;
{
	Cardinal i, j, k;
	Boolean file_exists = FALSE;
	
	j = strlen(name);

	for (i = 0; i < j; i++) {
		if (name[i] == '/' ||
		    isspace(name[i]) || !isprint(name[i]))
			name[i] = '.';
	}

	if (butt->save_directory != NULL)
		(void)sprintf(full_path, "%s/%s",
				butt->save_directory, name);
	else
		(void)strcpy(full_path, name);

	do {
		if (access(full_path, F_OK) == 0) {
			char *check;
			file_exists = TRUE;
			check = full_path + strlen(full_path) - 1;
			if (isdigit(*check)) {
				(*check)++;
			} else {
				static tries = 0;

				if (tries++ > 50)
					abort(-50);

				check[1] = '0';
				check[2] = '\0';
			}
		} else if (errno != ENOENT) /* cannot get at file */
			return (NULL);
		else
			file_exists = FALSE;
	} while (file_exists);

	return (full_path);
}


static void
save_all_settings(fd, butt)
FILE *fd;
struct app_resources *butt;
{
	Cardinal i;
	int x, y;
	String str;

	for (i = 1; i <= butt->num_servers; i++) {
		str = (String)XukcGetObjectOnList(butt->server_id,
							i);
		fprintf(fd, "%%XServer: %s\n", str);
		str = (String)XukcGetObjectOnList(butt->geometry,
							i);
		fprintf(fd, "%%Position(%s)\n", str);
		str = (String)XukcGetObjectOnList(butt->foreground,
							i);
		if (str != NULL && *str != '\0')
			fprintf(fd, "%%%s: %s\n", XtNforeground, str);
		str = (String)XukcGetObjectOnList(butt->background,
							i);
		if (str != NULL && *str != '\0')
			fprintf(fd, "%%%s: %s\n", XtNbackground, str);
		str = (String)XukcGetObjectOnList(butt->font,
							i);
		if (str != NULL && *str != '\0')
			fprintf(fd, "%%%s: %s\n", XtNfont, str);
		str = (String)XukcGetObjectOnList(butt->edit_colour,
							i);
		if (str != NULL && *str != '\0')
			fprintf(fd, "%%%s: %s\n", XtNbarColour, str);
		str = (String)XukcGetObjectOnList(
				butt->highlight_colour, i);
		if (str != NULL && *str != '\0')
			fprintf(fd, "%%%s: %s\n", XtNhighlightColor, str);
		str = (String)XukcGetObjectOnList(butt->bitmap,
							i);
		if (str != NULL && *str != '\0')
			fprintf(fd, "%%%s: %s\n", XtNbitmap, str);
		str = (String)XukcGetObjectOnList(butt->left_bitmap,
							i);
		if (str != NULL && *str != '\0')
			fprintf(fd, "%%%s: %s\n", XtNleftBitmap, str);
		str = (String)XukcGetObjectOnList(butt->right_bitmap,
							i);
		if (str != NULL && *str != '\0')
			fprintf(fd, "%%%s: %s\n", XtNrightBitmap, str);
	}
}


static Cardinal
store_server_id(butt, server_str)
struct app_resources *butt;
String server_str;
{
	Cardinal i;

	/* search server list to see if it already exists .. in which case
	 * return its index number */
	for (i = 1; i <= butt->num_servers; i++) {
		String str = (String)XukcGetObjectOnList(
						butt->server_id, i);
		if (strcmp(str, server_str) == 0)
			/* found it */
			return (i);
	}
	/* didn't find server id so save this one in a new slot and
	 * put NULLs in the attribute slots */
	butt->num_servers++;
	(void)XukcAddObjectToList(&butt->geometry,
					(XtPointer)NULL, FALSE);
	(void)XukcAddObjectToList(&butt->foreground,
					(XtPointer)NULL, FALSE);
	(void)XukcAddObjectToList(&butt->background,
					(XtPointer)NULL, FALSE);
	(void)XukcAddObjectToList(&butt->font,
					(XtPointer)NULL, FALSE);
	(void)XukcAddObjectToList(&butt->edit_colour,
					(XtPointer)NULL, FALSE);
	(void)XukcAddObjectToList(&butt->highlight_colour,
					(XtPointer)NULL, FALSE);
	(void)XukcAddObjectToList(&butt->bitmap,
					(XtPointer)NULL, FALSE);
	(void)XukcAddObjectToList(&butt->left_bitmap,
					(XtPointer)NULL, FALSE);
	(void)XukcAddObjectToList(&butt->right_bitmap,
					(XtPointer)NULL, FALSE);
	return (XukcAddObjectToList(&butt->server_id,
				(XtPointer)server_str, FALSE));
}


static void
get_values_and_set(butt)
struct app_resources *butt;
{
#define CONVERT_SIZE 128
	char foreground[CONVERT_SIZE], background[CONVERT_SIZE];
	char font[CONVERT_SIZE];
	char bar_colour[CONVERT_SIZE], highlight_colour[CONVERT_SIZE];
	char bitmap[CONVERT_SIZE];
	char left_bitmap[CONVERT_SIZE], right_bitmap[CONVERT_SIZE];
	Cardinal server_id = butt->this_server;

	XtVaGetValues(butt->perly_button,
		   XtVaTypedArg, XtNforeground,
		   XtRString, foreground, CONVERT_SIZE, 
		   XtVaTypedArg, XtNbackground,
		   XtRString, background, CONVERT_SIZE, 
		   XtVaTypedArg, XtNfont, XtRString,
		   font, CONVERT_SIZE, 
		   XtVaTypedArg, XtNbarColour,
		   XtRString, bar_colour, CONVERT_SIZE, 
		   XtVaTypedArg, XtNhighlightColor,
		   XtRString, highlight_colour, CONVERT_SIZE, 
		   XtVaTypedArg, XtNbitmap, XtRString, bitmap,
		   CONVERT_SIZE, 
		   XtVaTypedArg, XtNleftBitmap,
		   XtRString, left_bitmap, CONVERT_SIZE, 
		   XtVaTypedArg, XtNrightBitmap,
		   XtRString, right_bitmap, CONVERT_SIZE, 
		   NULL);

	XtFree((String)XukcGetObjectOnList(butt->foreground, server_id));
	(void)XukcChangeObjectOnList(butt->foreground, server_id,
					(XtPointer)XtNewString(foreground));

	XtFree((String)XukcGetObjectOnList(butt->background, server_id));
	(void)XukcChangeObjectOnList(butt->background, server_id,
					(XtPointer)XtNewString(background));

	XtFree((String)XukcGetObjectOnList(butt->font, server_id));
	(void)XukcChangeObjectOnList(butt->font, server_id,
					(XtPointer)XtNewString(font));

	XtFree((String)XukcGetObjectOnList(butt->edit_colour, server_id));
	(void)XukcChangeObjectOnList(butt->edit_colour, server_id,
					(XtPointer)XtNewString(bar_colour));

	XtFree((String)XukcGetObjectOnList(butt->highlight_colour, server_id));
	(void)XukcChangeObjectOnList(butt->highlight_colour, server_id,
				(XtPointer)XtNewString(highlight_colour));

	XtFree((String)XukcGetObjectOnList(butt->bitmap, server_id));
	(void)XukcChangeObjectOnList(butt->bitmap, server_id,
					(XtPointer)XtNewString(bitmap));

	XtFree((String)XukcGetObjectOnList(butt->left_bitmap, server_id));
	(void)XukcChangeObjectOnList(butt->left_bitmap, server_id,
					(XtPointer)XtNewString(left_bitmap));

	XtFree((String)XukcGetObjectOnList(butt->right_bitmap, server_id));
	(void)XukcChangeObjectOnList(butt->right_bitmap, server_id,
					(XtPointer)XtNewString(right_bitmap));
}


static void
save_button(fd, butt, user_invoked)
FILE *fd;
struct app_resources *butt;
Boolean user_invoked;
{
	Dimension internal_width, internal_height;
	Dimension highlight_thickness;
	String name, script;

	if (butt->this_server == NULL)
		butt->this_server = store_server_id(butt, get_server_id(butt));

	if (user_invoked) {
		StoreGeometry(butt);
		get_values_and_set(butt);
	}
	XtVaGetValues(butt->perly_button, XtNlabel, &name,
			   XtNhighlightThickness, &highlight_thickness, 
			   XtNinternalWidth, &internal_width,
			   XtNinternalHeight, &internal_height, NULL);

	if (butt->edit_popup != NULL)
		script = XukcGetTextWidgetString(butt->edit_script);
	else
		script = butt->script;

	fprintf(fd, "#!%s\n", PROGRAM_NAME);
	fprintf(fd, "%% Perly Button Description File\n");
	fprintf(fd, "%%ButtonName: %s\n", name);
	save_all_settings(fd, butt);
	fprintf(fd, "%%highlightThickness: %d\n", (int)highlight_thickness);
	fprintf(fd, "%%internalWidth: %d\n", (int)internal_width);
	fprintf(fd, "%%internalHeight: %d\n", (int)internal_height);
	if (butt->execution_directory != NULL)
		fprintf(fd, "%%*executionDirectory: %s\n",
				butt->execution_directory);
	fprintf(fd, "%%*beepWhenFinished: %s\n", butt->beep ? "true" : "false");
	fprintf(fd, "%%*popupResponse: %s\n",
		butt->popup_response == POPUP_ON_OUTPUT ?
		XukcEpopupOnOutput :
		butt->popup_response == POPUP_ON_COMPLETION ?
		XukcEpopupOnCompletion : XukcEnoPopup);
	fprintf(fd, "%%*iconResponse: %s\n",
		butt->icon_response == ICON_ON_WHEN_OUTPUT ?
		XukcEiconOnWhenOutput :
		butt->icon_response == ICON_ALWAYS_ON ?
		XukcEiconAlwaysOn : XukcEiconAlwaysOff);
	fprintf(fd, "%%*showQuit: %s\n", butt->show_quit ? "true" : "false");
	fprintf(fd, "%%*passSelection: %s\n", butt->pass_selection ?
						"true" : "false");
	fprintf(fd, "%%*splitSelection: %s\n", butt->split_selection ?
						"true" : "false");
	fprintf(fd, "%%*useStdin: %s\n", butt->use_stdin ? "true" : "false");
	fprintf(fd, "%%*autoClear: %s\n", butt->auto_clear ? "true" : "false");
	fprintf(fd, "%%*popdown: %s\n", butt->popdown_on_completion
						? "true" : "false");
	fprintf(fd, "%%SCRIPT:\n");
	if (script != NULL)
		fprintf(fd, "%s", script);
}


static Boolean
load_button(fd, butt, file)
FILE *fd;
struct app_resources *butt;
String file;
{
#define SET_ATTRIBUTE(name, list) \
	tmp = (String)XukcGetObjectOnList(list, butt->this_server); \
	if (tmp != NULL) /* override the default */ \
		XtVaSetValues(butt->perly_button, XtVaTypedArg, \
				name, XtRString, tmp, strlen(tmp) + 1, NULL); \
	else { \
		XtVaGetValues(butt->perly_button, XtVaTypedArg, \
				name, XtRString, input_str, 999, NULL); \
		(void)StoreAttribute(butt, name, input_str); \
	}
#define GETCHAR ch = fgetc(fd);
	String tmp;
	int ch;
	char input_str[1000];
	char input_value[1000];
	int a, b;
	Position x, y;
	Dimension width, height;
	Boolean found_app_resource;
	XrmDatabase xrdb;
	Cardinal server_id = 0;
	Cardinal i;

	found_app_resource = FALSE;
	GETCHAR;
	/* skip the first comment lines */
	while (ch == '#') {
		while (ch != EOF && ch != '\n')
			GETCHAR;
		GETCHAR;
	}
	(void)fscanf(fd, " Perly Button Description File\n");
	if (fscanf(fd, "%%ButtonName: %[^\n]\n", input_str) != 1) {
		char message[1000];

		(void)sprintf(message, "%s not a Perly Button Description File",
					file);
		XtWarning(message);
		return FALSE;
	}

	xrdb = XtDatabase(XtDisplay(butt->perly));
	XtVaSetValues(butt->perly_button, XtNlabel, input_str, NULL);

	butt->this_server = 0;
	GETCHAR;
	ungetc(ch, fd);
	while (ch != EOF) {
		GETCHAR; /* should be '%' */

		if (fscanf(fd, "XServer: %[^\n]\n", input_str) == 1) {
			butt->this_server = store_server_id(butt,
						XtNewString(input_str));
			fscanf(fd, "%%Position(%[^)])", input_str);
			if (!StoreAttribute(butt, XtNgeometry, input_str)) {
				(void)fprintf(stderr, "perly: internal corruption found whilst reading file %s\n",
				file);
				abort(-1);
			}
		
			GETCHAR;
			while (ch != '\n')
				GETCHAR;
			ungetc(ch, fd);
		} else if (ch == '%') { /* start of a resource line */
			input_str[0] = '\0';
			input_value[0] = '\0';
			fscanf(fd, "%[^:]: %[^\n]", input_str, input_value);
				found_app_resource = (*input_str == '*');

			if (strcmp(input_str, "SCRIPT") == 0) {
				/* found the script itself */
				butt->script = XtNewString(input_value);
				do {
					ch = fread(input_str, sizeof(char),
								999, fd);
					if (ch > 0) {
						input_str[ch] = '\0';
						butt->script =
							XtRealloc(butt->script,
							   strlen(butt->script)
							          + ch + 1);
						strcat(butt->script, input_str);
					}
				} while (ch > 0);
			} else if (found_app_resource)
				/* application resources */
				XrmPutStringResource(&xrdb, input_str,
								input_value);
			else if (!StoreAttribute(butt, input_str, input_value))
				/* found a widget resource line */
				XtVaSetValues(butt->perly_button, XtVaTypedArg,
						input_str, XtRString,
						input_value,
						strlen(input_value) + 1,
						NULL);
		}
		GETCHAR; /* get a newline or EOF */
	}

	/* retrieve all the attribute values for this server */
	butt->this_server = store_server_id(butt, get_server_id(butt));

	tmp = (String)XukcGetObjectOnList(butt->geometry, butt->this_server);
	if (tmp != NULL) /* override the default */
		XtVaSetValues(butt->perly, XtNgeometry, tmp, NULL);
	SET_ATTRIBUTE(XtNforeground, butt->foreground);
	SET_ATTRIBUTE(XtNbackground, butt->background);
	SET_ATTRIBUTE(XtNfont, butt->font);
	SET_ATTRIBUTE(XtNbarColour, butt->edit_colour);
	SET_ATTRIBUTE(XtNhighlightColor, butt->highlight_colour);
	SET_ATTRIBUTE(XtNbitmap, butt->bitmap);
	SET_ATTRIBUTE(XtNleftBitmap, butt->left_bitmap);
	SET_ATTRIBUTE(XtNrightBitmap, butt->right_bitmap);

	XtVaGetValues(butt->perly_button, XtNwidth, &width,
						XtNheight, &height, NULL);
	XtVaSetValues(butt->perly, XtNwidth, width, XtNheight, height, NULL);
	return TRUE;
}


/*************** Module Exported Routines *******************/


void
Save(butt, user_invoked)
struct app_resources *butt;
Boolean user_invoked;
{
	static char full_path[1000];
	FILE *fd;

	if (butt->filename == NULL) {
		String name;

		XtVaGetValues(butt->perly_button, XtNlabel, &name, NULL);
		name = XtNewString(name);
		butt->filename = cvt_to_filename(butt, full_path, name);
		XtFree(name);
	} else
		(void)strcpy(full_path, butt->filename);

	if (butt->filename == NULL || (fd = fopen(full_path, "w")) == NULL) {
		char message[1000];

		(void)sprintf(message, "Cannot save button to file: %s (%s)",
					full_path, reason());
		XtWarning(message);
		return;
	}

	save_button(fd, butt, user_invoked);

	if (chmod(full_path, 0755) < 0) {
		char message[1000];

                (void)sprintf(message, "Cannot chmod button file: %s (%s)",
                                        full_path, reason());
                XtWarning(message);
	}

	if (fclose(fd) < 0) {
		char message[1000];

		(void)sprintf(message,
			"Could not complete saving of button to file: %s (%s)",
			full_path, reason());
		XtWarning(message);
		return;
	}
}


Boolean
Load(butt, file)
struct app_resources *butt;
String file;
{
	FILE *fd;
	Boolean result;

	if (file == NULL)
		return TRUE;

	if (strcmp(file, "stdin") == 0)
		fd = stdin;
	else if ((fd = fopen(file, "r")) == NULL) {
		/* look for file in PERLY_BUTTONS directory */
		char *env_sd = getenv(SAVE_DIRECTORY);
		String full_path = NULL;

		if (env_sd != NULL) {
			full_path = XtMalloc(strlen(env_sd) + strlen(file) + 2);
			(void)sprintf(full_path, "%s/%s", env_sd, file);
		}
			
		if (full_path == NULL || (fd = fopen(full_path, "r")) == NULL) {
			char message[1000];

			(void)sprintf(message,
					"Cannot load button from file: %s (%s)",
					file, reason());
			XtWarning(message);
			if (full_path != NULL)
				XtFree(full_path);
			return FALSE;
		}
		butt->filename = full_path;
	}

	result = load_button(fd, butt, file);

	if (fd != stdin && fclose(fd) < 0) {
		char message[1000];

		(void)sprintf(message,
			"Could not complete loading of button from file: %s (%s)",
			file, reason());
		XtWarning(message);
	}
	return result;
}


void
RemoveFile(butt)
struct app_resources *butt;
{
	(void)unlink(butt->filename);
}


void
InitializeFileSaving(butt, file)
struct app_resources *butt;
String file;
{
	if (butt->filename == NULL) {
		if (file != NULL)
			butt->filename = XtNewString(file);
		else
			butt->filename = NULL;
	}

	if (butt->save_directory == NULL) {
		/* look for a environment variable */
		char *env_sd = getenv(SAVE_DIRECTORY);

		if (env_sd == NULL)
			butt->save_directory = ".";
		else
			butt->save_directory = env_sd;
	}
}


Boolean
StoreAttribute(butt, type, str)
struct app_resources *butt;
String type;
String str;
{
	String old_str;
	Cardinal *list;

	if (strcmp(type, XtNforeground) == 0)
		list = &butt->foreground;
	else if (strcmp(type, XtNbackground) == 0)
		list = &butt->background;
	else if (strcmp(type, XtNfont) == 0)
		list = &butt->font;
	else if (strcmp(type, XtNbarColour) == 0)
		list = &butt->edit_colour;
	else if (strcmp(type, XtNhighlightColor) == 0)
		list = &butt->highlight_colour;
	else if (strcmp(type, XtNbitmap) == 0)
		list = &butt->bitmap;
	else if (strcmp(type, XtNleftBitmap) == 0)
		list = &butt->left_bitmap;
	else if (strcmp(type, XtNrightBitmap) == 0)
		list = &butt->right_bitmap;
	else if (strcmp(type, XtNgeometry) == 0)
		list = &butt->geometry;
	else
		return FALSE;

	if (*list == 0 || butt->this_server == NULL)
		butt->this_server = store_server_id(butt, get_server_id(butt));
		
	old_str = (String)XukcGetObjectOnList(*list, butt->this_server);
	XtFree(old_str);
	(void)XukcChangeObjectOnList(*list, butt->this_server,
					(XtPointer)(str == NULL ? NULL :
							XtNewString(str)));
	return TRUE;
}


void
CopyAndStoreAttributes(from, to)
struct app_resources *from, *to;
{
	/* copies over all the string based attributes for this server */
	String str;
#define COPY(list) \
	str = (String)XukcGetObjectOnList(from->list, from->this_server); \
	if (str != NULL) { \
		XtFree((String)XukcGetObjectOnList(to->list, to->this_server)); \
		(void)XukcChangeObjectOnList(to->list, to->this_server, \
					(XtPointer)XtNewString(str)); \
	}

	COPY(foreground);
	COPY(background);
	COPY(font);
	COPY(edit_colour);
	COPY(highlight_colour);
	COPY(bitmap);
	COPY(left_bitmap);
	COPY(right_bitmap);
}


void
SetInitializeAttributes(butt)
struct app_resources *butt;
{
	butt->this_server = store_server_id(butt, get_server_id(butt));
	get_values_and_set(butt);
}


void
StoreGeometry(butt)
struct app_resources *butt;
{
	char geom[50];
	Position x = 0, y = 0;
	Dimension junk;
#if 0
	char user_geom[100];
	char def_geom[100];
	unsigned int bwidth;
	XSizeHints hints;
	int x_ret, y_ret, junk1;
#endif

	XFlush(XtDisplay(butt->perly));
#if 1
	if (XukcFindUndecoratedWindow(XtDisplay(butt->perly),
	    XtWindow(butt->perly), &x, &y, &junk, &junk, &junk) == NULL)
		return;  /* an Xlib call failed .. so ignore it */

	/* wm_offset is a horrible kludge to fix MWM 1.1 which gets
	 * window geometries wrong by +2+2 pixels! yuk */
	set_wm_offset(XtDisplay(butt->perly));
	(void)sprintf(geom, "+%d+%d", x - wm_offset, y - wm_offset);

	/* stops a bug? */
	if (x == 0 && y == 0)
		return;
#else
	/* ||| */
	if (!XWMGeometry(XtDisplay(butt->perly), 0, user_geom, def_geom,
	    &bwidth, &hints, &x_ret, &y_ret, &junk1, &junk1, &junk1))
		return;

	(void)sprintf(geom, "+%d+%d", x_ret, y_ret);

	/* stops a bug? */
	if (x_ret == 0 && y_ret == 0)
		return;
#endif

	(void)StoreAttribute(butt, XtNgeometry, geom);
}
