#ifndef lint
static char rcsid[] =
   "$Xukc: perly.c,v 1.8 91/04/15 18:57:27 rlh2 Exp $";
#endif /* !lint */

/* 
 * Copyright 1990 Richard Hesketh / rlh2@ukc.ac.uk
 *                Computing Lab. University of Kent at Canterbury, UK
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of Richard Hesketh and The University of
 * Kent at Canterbury not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Richard Hesketh and The University of Kent at Canterbury make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * Richard Hesketh AND THE UNIVERSITY OF KENT AT CANTERBURY DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL Richard Hesketh OR THE
 * UNIVERSITY OF KENT AT CANTERBURY BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 *
 * Author:  Richard Hesketh / rlh2@ukc.ac.uk, 
 *                Computing Lab. University of Kent at Canterbury, UK
 */

#include <ctype.h>
#include <signal.h>
#include <X11/Xatom.h>
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xukc/Command.h>
#include <X11/Xukc/text.h>
#include <X11/Xukc/memory.h>
#include <X11/Xukc/Convert.h>
#include "perly.h"

#define SWAP(a,b) { Position c; c = a; a = b; b = c; }
#define START "start"
#define MOVE "move"
#define STOP "stop"
#define RESIZE "resize"
#define COPY "copy"
#define NONE 0
#define STARTED 1
#define MOVING 2
#define TIMEOUT 1000

struct app_resources *GetCurrentButton();

/* perly_create.c */
extern Widget XukcDirtGO();

/* execute.c */
extern void SendSignal();
extern int RunScript();
extern void SetIconWidth();
extern void CleanUpOnChildEOF();

/* tailor.c */
extern void InitializeTailoring();
extern struct app_resources *CurrentTailoring();

/* file.c */
extern void InitializeFileSaving();
extern Boolean Load();
extern void StoreGeometry();
extern void CopyAndStoreAttributes();
extern void SetInitializeAttributes();
extern void Save();
extern void RemoveFile();
extern void SetDirectory();

extern Widget WI_tailor_popup, WI_hide_tailor_button;

void PositionPopups();
void BindOutputInterface();
void create_and_initialize_button();

struct app_resources *current_button = NULL;
struct app_resources *creation_button = NULL;

static Cardinal button_list = 0;

static Atom perly_master_atom = NULL, perly_new_atom = NULL;

static void perly_popup();
void TailoringPopup();
static void quit_perly();
static void save_perly();
static void perly_move();
static void finished_executing();
static void change_perly_sensitivity();

static XtActionsRec actions[] = {
	{ "perly_popup", perly_popup },
	{ "tailor_popup", TailoringPopup },
	{ "perly_move", perly_move },
	{ "save_perly", save_perly },
	{ "quit_perly", quit_perly },
};

static XrmOptionDescRec options[] = {
	{ "-popup", ".popupResponse", XrmoptionSepArg, (caddr_t)NULL },
	{ "-icon", ".iconResponse", XrmoptionSepArg, (caddr_t)NULL },
	{ "-execdir", ".executionDirectory", XrmoptionSepArg, (caddr_t)NULL },
	{ "-savedir", ".saveDirectory", XrmoptionSepArg, (caddr_t)NULL },
	{ "-noargs", ".passSelection", XrmoptionNoArg, (caddr_t)"false" },
	{ "-nosplit", ".splitSelection", XrmoptionNoArg, (caddr_t)"false" },
	{ "-quit", ".showQuit", XrmoptionNoArg, (caddr_t)"true" },
	{ "+beep", ".beepWhenFinished", XrmoptionNoArg, (caddr_t)"true" },
	{ "+stdin", ".useStdin", XrmoptionNoArg, (caddr_t)"true" },
	{ "+clear", ".autoClear", XrmoptionNoArg, (caddr_t)"true" },
	{ "+popdown", ".popdown", XrmoptionNoArg, (caddr_t)"true" },
};

#define offset(field) XtOffset(struct app_resources *, field)
static XtResource perly_resources[] = {
	{ "iconResponse", "IconResponse", XukcRIconResponse, sizeof(int),
	  offset(icon_response), XtRImmediate, (XtPointer)ICON_ON_WHEN_OUTPUT },
	{ "popupResponse", "PopupResponse", XukcRPopupResponse, sizeof(int),
	  offset(popup_response), XtRImmediate, (XtPointer)POPUP_ON_OUTPUT },
	{ "executionDirectory", "ExecutionDirectory", XtRString, sizeof(String),
	  offset(execution_directory), XtRString, (XtPointer)NULL },
	{ "saveDirectory", "SaveDirectory", XtRString, sizeof(String),
	  offset(save_directory), XtRString, (XtPointer)NULL },
	{ "passSelection", "PassSelection", XtRBoolean, sizeof(Boolean),
	  offset(pass_selection), XtRImmediate, (XtPointer)TRUE },
	{ "splitSelection", "SplitSelection", XtRBoolean, sizeof(Boolean),
	  offset(split_selection), XtRImmediate, (XtPointer)TRUE },
	{ "showQuit", "ShowQuit", XtRBoolean, sizeof(Boolean),
	  offset(show_quit), XtRImmediate, (XtPointer)FALSE },
	{ "beepWhenFinished", "BeepWhenFinished", XtRBoolean, sizeof(Boolean),
	  offset(beep), XtRImmediate, (XtPointer)FALSE },
	{ "useStdin", "UseStdin", XtRBoolean, sizeof(Boolean),
	  offset(use_stdin), XtRImmediate, (XtPointer)FALSE },
	{ "autoClear", "AutoClear", XtRBoolean, sizeof(Boolean),
	  offset(auto_clear), XtRImmediate, (XtPointer)FALSE },
	{ "popdown", "Popdown", XtRBoolean, sizeof(Boolean),
	  offset(popdown_on_completion), XtRImmediate, (XtPointer)FALSE },
};

#define NEW_CONVERTER
#ifdef NEW_CONVERTER
static EnumListRec popup_response[] = {
	{ XukcEpopupOnOutput, (int)POPUP_ON_OUTPUT },
	{ XukcEpopupOnCompletion, (int)POPUP_ON_COMPLETION },
	{ XukcEnoPopup, (int)NO_POPUP },
};

static XtConvertArgRec popup_response_Args[] = {
	{ XtAddress,	(XtPointer)popup_response, sizeof(EnumList) },
	{ XtImmediate,	(XtPointer)XtNumber(popup_response), sizeof(Cardinal) },
	{ XtAddress,	XukcRPopupResponse, sizeof(String) },

};

static EnumListRec icon_response[] = {
	{ XukcEiconOnWhenOutput, (int)ICON_ON_WHEN_OUTPUT },
	{ XukcEiconAlwaysOn, (int)ICON_ALWAYS_ON },
	{ XukcEiconAlwaysOff, (int)ICON_ALWAYS_OFF },
};

static XtConvertArgRec icon_response_Args[] = {
	{ XtAddress,	(XtPointer)icon_response, sizeof(EnumList) },
	{ XtImmediate,	(XtPointer)XtNumber(icon_response), sizeof(Cardinal) },
	{ XtAddress,	XukcRIconResponse, sizeof(String) },
};
#else
static QIListRec popup_response[] = {
	{ (XrmQuark)NULL, XukcEpopupOnOutput, (int)POPUP_ON_OUTPUT },
	{ (XrmQuark)NULL, XukcEpopupOnCompletion, (int)POPUP_ON_COMPLETION },
	{ (XrmQuark)NULL, XukcEnoPopup, (int)NO_POPUP },
};

static XtConvertArgRec popup_response_Args[] = {
	{ XtAddress,	(XtPointer)popup_response, sizeof(QIList) },
	{ XtImmediate,	(XtPointer)XtNumber(popup_response), sizeof(Cardinal) },
	{ XtAddress,	XukcRPopupResponse, sizeof(String) },

};

static QIListRec icon_response[] = {
	{ (XrmQuark)NULL, XukcEiconOnWhenOutput, (int)ICON_ON_WHEN_OUTPUT },
	{ (XrmQuark)NULL, XukcEiconAlwaysOn, (int)ICON_ALWAYS_ON },
	{ (XrmQuark)NULL, XukcEiconAlwaysOff, (int)ICON_ALWAYS_OFF },
};

static XtConvertArgRec icon_response_Args[] = {
	{ XtAddress,	(XtPointer)icon_response, sizeof(QIList) },
	{ XtImmediate,	(XtPointer)XtNumber(icon_response), sizeof(Cardinal) },
	{ XtAddress,	XukcRIconResponse, sizeof(String) },
};
#endif /* NEW_CONVERTER */

static XSegment lines[8]; 
static Boolean doing_selection = FALSE;
static String current_selection = NULL;

static Atom wm_save_yourself = NULL;
static Atom wm_protocols = NULL;

static void
update_wm_command()
{
	struct app_resources **buttons;
	Cardinal i, j, num_buttons;
	char **command_line;

	/* build a command line to invoke perly and create all the current
	 * on screen buttons.
	 */
	if (button_list == 0 || !XtIsRealized(creation_button->perly))
		return;

	buttons = (struct app_resources **)XukcGetListPointer(button_list,
								&num_buttons);
	
	command_line = (char **)XtMalloc((num_buttons + 1) * sizeof(char *));
	command_line[0] = PROG_NAME;
	/* ||| should also save real command line arguments??? ||| */
	for (i = j = 0; i < num_buttons; i++)
		if (buttons[i]->filename) {
			if (buttons[i]->save_directory != NULL &&
			    strncmp(buttons[i]->save_directory,
				    buttons[i]->filename,
				    strlen(buttons[i]->save_directory)) == 0)
				command_line[i+1] = buttons[i]->filename +
					strlen(buttons[i]->save_directory)
					+ 1;
			else
				command_line[i+1] = buttons[i]->filename;
			j++;
		}

	XSync(XtDisplay(creation_button->perly), FALSE);
	XSetCommand(XtDisplay(creation_button->perly),
		    XtWindow(creation_button->perly),
		    command_line, j + 1);
	XtFree((char *)command_line);
}


static XtEventHandler
process_client_message(w, closure, event, cont)
Widget w;
XtPointer closure;
XEvent *event;
Boolean *cont;
{
	/* we get here when a ClientMessage event gets sent to the
	 * top-level window and so we check to see whether it is a
	 * "WM_SAVE_YOURSELF" message in which case we update the
	 * "WM_COMMAND" property.
	 */
	if (event->xclient.window == XtWindow(w) &&
	    event->xclient.message_type == wm_protocols &&
	    event->xclient.data.l[0] == wm_save_yourself) {
		update_wm_command();
	}
}


static void
set_wm_protocols(w)
Widget w;
{
	/* create a WM_PROTOCOLS property on the top window and
	 * say that we respond to WM_SAVE_YOURSELF to update the
	 * WM_COMMAND .. useful for things like my "snapshot" program.
	 */
	wm_protocols = XInternAtom(XtDisplay(w), "WM_PROTOCOLS", FALSE);
	wm_save_yourself = XInternAtom(XtDisplay(w), "WM_SAVE_YOURSELF", FALSE);
	(void)XSetWMProtocols(XtDisplay(w), XtWindow(w), &wm_save_yourself, 1);

	XtAddEventHandler(w, NoEventMask, TRUE, process_client_message,
				(XtPointer)NULL);
}


static int
signal_save()
{
	struct app_resources **buttons;
	Cardinal num_buttons;

	/* the process is dieing due to some external force so save
	 * everything insight! - "flee run for your lives the muncher is
	 * coming!"
	 */
	buttons = (struct app_resources **)XukcGetListPointer(button_list,
								&num_buttons);
	if (num_buttons == 0)
		exit(0);

	for (; num_buttons > 0; num_buttons--) {
		Save(buttons[num_buttons-1], FALSE);
		if (buttons[num_buttons-1]->scriptfile != NULL)
			(void)unlink(buttons[num_buttons-1]->scriptfile);
	}
	exit(0);
}

static Widget
toplevel_parent(w)
Widget w;
{
	while (XtParent(w) != NULL)
		w = XtParent(w);
	return (w);
}

/********************* Resize/Move Frame *************************/

static void
calculate_end_points(w, resize, ev, ax, ay, bx, by)
Widget w;
Boolean resize;
XEvent *ev;
Position *ax, *ay, *bx, *by;
{
	static Position sx, sy;
	Position x, y;
	Dimension width, height;

	if (ev->type == MotionNotify) {
		x = ev->xmotion.x;
		y = ev->xmotion.y;
	} else {
		sx = x = ev->xbutton.x;
		sx = y = ev->xbutton.y;
	}

	XtVaGetValues(w, XtNwidth, &width, XtNheight, &height, NULL);

	if (!resize) {
#if 0
		*ax = x - sx; *ay = y - sy;
		*bx = *ax + width; *by = *ay + height;
#else
		*ax = x - width/2; *ay = y - height/2;
		*bx = *ax + width; *by = *ay + height;
#endif
	} else {
		*ax = x; *ay = y; *bx = 0; *by = 0;
	} 

	if (*ax > *bx)
		SWAP(*ax, *bx);
	if (*ay > *by)
		SWAP(*ay, *by);
}


/* creates a wire frame between two extremes.  Extremes must be opposite 
 * corners in a rectangle.  Updates the lines[] array. */
static void
make_frame(w, ax, ay, bx, by)
Widget w;
Position ax, ay, bx, by;
{
	Position x1, y1, x2, y2;

	/* we are drawing into the root window so translate the coords */
	XtTranslateCoords(w, ax, ay, &x1, &y1);
	XtTranslateCoords(w, bx, by, &x2, &y2);

	/* x1, y1 to x2, y2 defines a box (0,0)-(a,b) */
	lines[0].x1 = x1; lines[0].y1 = y1;  /* (0,0)-(a,0) */
	lines[0].x2 = x2; lines[0].y2 = lines[0].y1;
	lines[1].x1 = x2; lines[1].y1 = y1;  /* (a,0)-(a,b) */
	lines[1].x2 = x2; lines[1].y2 = y2;
	lines[2].x1 = x2; lines[2].y1 = y2;  /* (a,b)-(0,b) */
	lines[2].x2 = x1; lines[2].y2 = lines[2].y1;
	lines[3].x1 = x1; lines[3].y1 = y2;  /* (0,b)-(0,0) */
	lines[3].x2 = x1; lines[3].y2 = y1;
	lines[4].x1 = x1; lines[4].y1 = (y2-y1)/3+y1;  /* (0,b/3)-(a,b/3) */
	lines[4].x2 = x2; lines[4].y2 = lines[4].y1;
	lines[5].x1 = x1; lines[5].y1 = (y2-y1)*2/3+y1;/* (0,b*2/3)-(a,b*2/3) */
	lines[5].x2 = x2; lines[5].y2 = lines[5].y1;
	lines[6].x1 = (x2-x1)/3+x1; lines[6].y1 = y1;  /* (a/3,0)-(a/3,b) */
	lines[6].x2 = lines[6].x1; lines[6].y2 = y2;
	lines[7].x1 = (x2-x1)*2/3+x1; lines[7].y1 = y1;/* (a*2/3,0)-(a*2/3,b) */
	lines[7].x2 = lines[7].x1; lines[7].y2 = y2;
}


static void
draw_frame(w)
Widget w;
{
	static GC invert_gc = NULL;
	XGCValues gcvalues;

	if (invert_gc == NULL) {
		/* make the GC that Xors the lines of a frame */
		gcvalues.function = GXinvert;
		gcvalues.line_width = 0;
		gcvalues.subwindow_mode = IncludeInferiors;
		invert_gc = XtGetGC(w, GCFunction | GCLineWidth |
						GCSubwindowMode, &gcvalues);
	}
	
	XDrawSegments(XtDisplay(w), XtScreen(w)->root, invert_gc, lines, 8);
}


/* drag resize a corner or side.  uses a skeleton frame to show where
 * the new corner will be and then resizes and moves the target widget.
 * This action is called from the perly button and can therefore
 * have 3 states, "just picked up (press)", "dragged (move)" and "just let go
 * (release)", which starts, moves and removes the dragged lines.  The
 * widget is resized and moved when the "release" is set.
 */
static void
perly_move(w, event, params, num_params)
Widget w;
XEvent *event;
String *params;
Cardinal *num_params;
{
	static Cardinal state = NONE;
	Position ax, ay, bx, by;
	Position x, y;
	Boolean doing_resize, doing_copy;

	if (*num_params != 2)
		XtError("resize_target() called with incorrect parameter count");

	doing_resize = (strcmp(params[1], RESIZE) == 0);
	doing_copy = (strcmp(params[1], COPY) == 0);

	if (strcmp(params[0], START) == 0 && state == NONE) {
		/* start tracking the cursor and drawing the frame */
		XGrabServer(XtDisplay(w));
		calculate_end_points(w, doing_resize, event,
							&ax, &ay, &bx, &by);
		make_frame(w, ax, ay, bx, by);
		draw_frame(w);
		state = STARTED;
	} else if (strcmp(params[0], MOVE) == 0 && state != NONE) {
		/* move the frame to the new tracked cursor position */
		calculate_end_points(w, doing_resize, event,
							&ax, &ay, &bx, &by);
		draw_frame(w);
		make_frame(w, ax, ay, bx, by);
		draw_frame(w);
		state = MOVING;
	} else if (strcmp(params[0], STOP) == 0) {
		/* finished resizing or moving so update the target widget */
		draw_frame(w);
		calculate_end_points(w, doing_resize, event,
							&ax, &ay, &bx, &by);
		XUngrabServer(XtDisplay(w));
		state = NONE;   
		XtVaGetValues(toplevel_parent(w), XtNx, &x, XtNy, &y, NULL);
		x += ax;
		y += ay;

		if (doing_resize) {
			/* update the target, if the geometry is allowed
			 * the change will be made automatically. */
			Dimension width, height;

			width = bx - ax;
			height = by - ay;
			XtVaSetValues(w, XtNwidth, width,
					XtNheight, height, NULL);
			XtVaSetValues(toplevel_parent(w),
						XtNx, x, XtNy, y, NULL);
			StoreGeometry(GetCurrentButton(w));
		} else {
			if (doing_copy) {
				create_and_initialize_button(x, y,
						GetCurrentButton(w),
						(String)NULL);
			} else {
				XtVaSetValues(toplevel_parent(w),
							XtNx, x, XtNy, y, NULL);
				StoreGeometry(GetCurrentButton(w));
			}
		}
	}
}

/************************ End of Resize/Move Frame *********************/

/************************ Selection Mechanism **************************/

static void
selection_done(w, client_data, selection, type, value, length, format)
Widget w;
XtPointer client_data;
Atom *selection;
Atom *type;
XtPointer value;
unsigned long *length;
int *format;
{
	if (*type == XT_CONVERT_FAIL)
		current_selection = NULL;
	else
		current_selection = (String)value;
	doing_selection = FALSE;
}


static char **
get_selected_arguments(w, split, num_args, list_num)
Widget w;
Boolean split;
int *num_args;
Cardinal *list_num;
{
	Cardinal i;
	char **args = NULL;
	char *s, *n;

	if (current_selection)
		XtFree(current_selection);
	*num_args = 0;
	*list_num = 0;
	doing_selection = TRUE;
	current_selection = NULL;

	XtGetSelectionValue(w, XA_PRIMARY, XA_STRING, selection_done,
					(XtPointer)NULL,
					XtLastTimestampProcessed(XtDisplay(w)));
	/* wait until the selection has been done or timed out */
	while (doing_selection) {
		XEvent event;

		XtAppNextEvent(XtWidgetToApplicationContext(w), &event);
		(void)XtDispatchEvent(&event);
	}

	if (current_selection == NULL)
		return (NULL);

	if (split) {
		/* split the selected string up into arguments using whitespace
		 * as argument delimiters.
		 */
		s = current_selection;
		while (*s != '\0') {
			while (*s && isspace(*s)) s++;
			if (*s) {
				n = s;
				while (*s && !isspace(*s)) s++;
				if (*s != '\0') {
					/* not reached end of selected string */
					*s = '\0';
					s++;
				}
				(void)XukcAddObjectToList(list_num, (XtPointer)n);
			}
		}
	} else
		(void)XukcAddObjectToList(list_num, (XtPointer)current_selection);

	args = (char **)XukcGetListPointer(*list_num, &i);
	*num_args = (int)i;
	return (args);
}

/******************** End of Selection Mechanism ***********************/

static void
perly_popup(w, event, params, num_params)
Widget w;
XEvent *event;
String *params;
Cardinal *num_params;
{
	current_button = GetCurrentButton(w);

	if (current_button->edit_is_up)
		XtPopdown(current_button->edit_popup);
	else {
		GetNewEditPopup(current_button);
		PositionPopups(current_button);
		XtPopup(current_button->edit_popup, XtGrabNone);
		XRaiseWindow(XtDisplay(current_button->edit_popup),
				XtWindow(current_button->edit_popup));
	}
	current_button->edit_is_up = !current_button->edit_is_up;
}


static void
perly_popdown(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	perly_popup(w, (XEvent *)NULL, (String *)NULL, (Cardinal *)NULL);
}


static void
tailor_popdown(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	TailoringPopup((Widget)NULL, (XEvent *)NULL, (String *)NULL, (Cardinal *)NULL);
}


static void
edit(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	current_button = (struct app_resources *)client_data;

	if ((Boolean)call_data)
		XtManageChild(current_button->edit_script);
	else
		XtUnmanageChild(current_button->edit_script);
}


static void
pause(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	current_button = (struct app_resources *)client_data;

	if ((Boolean)call_data)
		SendSignal(current_button, SIGSTOP);
	else
		SendSignal(current_button, SIGCONT);
}


static void
kill_timeout(client_data, id)
XtPointer client_data;
XtIntervalId *id;
{
	finished_executing((struct app_resources *)client_data);
}


static void
interrupt(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	struct app_resources *butt = (struct app_resources *)client_data;

	/* continue the process if on pause */
	XukcUnsetToggle(butt->pause_button, TRUE);
	/* .. and then send it the interrupt signal */
	SendSignal(butt, SIGCONT);
	SendSignal(butt, butt->sent_interrupt ? SIGKILL : SIGHUP);
	if (butt->sent_interrupt || butt->has_finished)
		/* we have sent a KILL but just incase set a timeout */
		(void)XtAppAddTimeOut(XtWidgetToApplicationContext(w),
						(unsigned long)TIMEOUT,
						kill_timeout,
						(XtPointer)butt);
	butt->sent_interrupt = TRUE;
}


static void
finished_executing(butt)
struct app_resources *butt;
{
	CleanUpOnChildEOF(butt);

	if (butt->output_pipe_id != NULL)
		XtRemoveInput(butt->output_pipe_id);
	if (butt->error_pipe_id != NULL)
		XtRemoveInput(butt->error_pipe_id);
	if (OUTPUT_POPUP_CREATED(butt)) {
		XtSetSensitive(butt->interrupt_button, FALSE);
		XtSetSensitive(butt->execute_button, TRUE);
		XtSetSensitive(butt->pause_button, FALSE);
		XukcUnsetToggle(butt->pause_button, FALSE);
	}
	XtSetSensitive(butt->perly_icon, FALSE);
	change_perly_sensitivity(butt, TRUE);
	butt->output_pipe_id = NULL;
	butt->error_pipe_id = NULL;
	if (butt->pipefds[0])
		close(butt->pipefds[0]);
	if (butt->pipefds[1])
		close(butt->pipefds[1]);
	butt->pipefds[0] = butt->pipefds[1] = NULL;

	if (butt->icon_response == ICON_ON_WHEN_OUTPUT)
		XtUnmanageChild(butt->perly_icon);

	if (!butt->edit_is_up && butt->popup_response == POPUP_ON_COMPLETION)
		perly_popup(butt->perly, (XEvent *)NULL, (String *)NULL,
							(Cardinal *)NULL);
	else if (butt->edit_is_up && butt->popdown_on_completion)
		perly_popup(butt->perly, (XEvent *)NULL, (String *)NULL,
							(Cardinal *)NULL);

	if (butt->beep)
		XBell(XtDisplay(butt->perly), 50);
	butt->is_executing = FALSE;
}


static void
update_icon(tw, butt)
Widget tw;
struct app_resources *butt;
{
	XawTextPosition start, end;
	XawTextBlock text;

	XawTextSetInsertionPoint(tw, (XawTextPosition)999999);
	end = XawTextGetInsertionPoint(tw);
	start = XawTextSourceScan(XawTextGetSource(tw), end, XawstEOL,
					XawsdLeft, NUM_ICON_LINES, TRUE);
	end = XawTextSourceRead(XawTextGetSource(tw), start, &text, end-start);
	XtVaSetValues(butt->perly_icon, XtNlabel, text.ptr, NULL);
	XtManageChild(butt->perly_icon);
	XRaiseWindow(XtDisplay(butt->perly), XtWindow(butt->perly));
}


/* output from the script is available on one of two pipes .. the stdout
 * pipe and the stderr pipe.   Drain the pipe a bit at a time unless the
 * script has finished in which case drain the whole of the pipe.
 * When both pipes have been drained they are closed and the script is
 * deemed to have stop executing and finished_executing() is called.
 */
static void
output_text(client_data, source, id)
XtPointer client_data;
int *source;
XtInputId *id;
{
#define READ_SIZE 512
	struct app_resources *butt = (struct app_resources *)client_data;
	char results[READ_SIZE+1];
	int num_read;

	do {
		results[0] = '\0';
		num_read = read(*source, results, READ_SIZE);
		if (num_read > 0) {
			results[num_read] = '\0';

			GetNewEditPopup(butt);
			if (*id == butt->output_pipe_id)
				XukcAppendToTextWidget(butt->stdout_text,
								results);
			else
				XukcAppendToTextWidget(butt->stderr_text,
								results);
			if (!butt->edit_is_up &&
			    butt->popup_response == POPUP_ON_OUTPUT)
				perly_popup(butt->perly, (XEvent *)NULL,
					(String *)NULL, (Cardinal *)NULL);
			if (butt->icon_response != ICON_ALWAYS_OFF)
				update_icon(*id == butt->output_pipe_id ?
						butt->stdout_text :
						butt->stderr_text,
						butt);
		} else {
			butt->has_finished = TRUE;
			XtRemoveInput(*id);
			if (*id == butt->output_pipe_id)
				butt->output_pipe_id = NULL;
			else
				butt->error_pipe_id = NULL;
		}
	} while (num_read > 0 && butt->has_finished);

	if (butt->has_finished) {
		if (butt->both_pipes_closed) {
			finished_executing(butt);
			butt->both_pipes_closed = FALSE;
		} else
			butt->both_pipes_closed = TRUE;
	}
}


static void
change_perly_sensitivity(butt, state)
struct app_resources *butt;
Boolean state;
{
	XtSetSensitive(butt->perly_button, state);

	if (!state) {
		Dimension width, height;
		String label;
		XWindowChanges values;

		XtVaGetValues(butt->perly_button, XtNlabel, &label,
				XtNwidth, &width, XtNheight, &height, NULL);
		XukcChangeLabel(butt->perly_button, label, TRUE);
		XtVaSetValues(butt->transparent, XtNbackgroundPixmap, None,
				XtNx, 0, XtNy, 0,
				XtNwidth, width, XtNheight, height, NULL);
		XtRealizeWidget(butt->transparent);
		XtMapWidget(butt->transparent);
		values.sibling = XtWindow(butt->perly_button);
		values.stack_mode = Above;
		XConfigureWindow(XtDisplay(butt->transparent),
					XtWindow(butt->transparent),
					 CWSibling | CWStackMode, &values);
		butt->just_mapped = TRUE;
	} else
		XtUnmapWidget(butt->transparent);
}


static void
exposure(w, client_data, event, cont)
Widget w;
XtPointer client_data;
XEvent *event;
Boolean *cont;
{
	struct app_resources *butt = (struct app_resources *)client_data;

	if (butt->just_mapped || !butt->is_executing) {
		butt->just_mapped = FALSE;
		return;
	}
	change_perly_sensitivity(butt, TRUE);
	change_perly_sensitivity(butt, FALSE);
}


static void
execute(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	char **args;
	int num_args = 0;
	String script;
	Cardinal list_num = 0;
	struct app_resources *butt = (struct app_resources *)client_data;

	if (butt->is_executing)
		return;

	change_perly_sensitivity(butt, FALSE);
	if (OUTPUT_POPUP_CREATED(butt))
		XtSetSensitive(butt->execute_button, FALSE);
	if (butt->pass_selection)
		args = get_selected_arguments(w, butt->split_selection,
						&num_args, &list_num);
	else {
		args = NULL;
		num_args = 0;
	}
	if (OUTPUT_POPUP_CREATED(butt)) {
		script = XukcGetTextWidgetString(butt->edit_script);
		XtFree(butt->script);
		butt->script = XtNewString(script);
	}
	butt->has_finished = FALSE;
	butt->output_pipe_id = NULL;
	butt->error_pipe_id = NULL;

	if (RunScript(butt, XtDisplay(w), butt->script, args, num_args)) {
		butt->both_pipes_closed = FALSE;
		butt->output_pipe_id = XtAppAddInput(
					XtWidgetToApplicationContext(w),
					butt->pipefds[0], XtInputReadMask,
					output_text, (XtPointer)butt);
		butt->error_pipe_id = XtAppAddInput(
					XtWidgetToApplicationContext(w),
					butt->pipefds[1], XtInputReadMask,
					output_text, (XtPointer)butt);
		XtSetSensitive(butt->perly_icon, TRUE);
		if (butt->icon_response != ICON_ALWAYS_OFF)
			XtVaSetValues(butt->perly_icon, XtNlabel, "", NULL);

		if (OUTPUT_POPUP_CREATED(butt)) {
			XtSetSensitive(butt->interrupt_button, TRUE);
			XtSetSensitive(butt->pause_button, TRUE);
			if (butt->auto_clear) {
				XtCallCallbacks(butt->stderr_clear,
						XtNcallback, (XtPointer)NULL);
				XtCallCallbacks(butt->stdout_clear,
						XtNcallback, (XtPointer)NULL);
			}
		}
			
		butt->sent_interrupt = FALSE;
		butt->is_executing = TRUE;
	} else {
		change_perly_sensitivity(butt, TRUE);
		if (OUTPUT_POPUP_CREATED(butt))
			XtSetSensitive(butt->execute_button, TRUE);
		XtSetSensitive(butt->perly_icon, FALSE);
		butt->is_executing = FALSE;
	}

	XukcDestroyList(&list_num, FALSE);
}


static void
clear(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	current_button = GetCurrentButton(w);
	XukcClearTextWidget((Widget)client_data);

	if (current_button->icon_response != ICON_ALWAYS_OFF)
		XtVaSetValues(current_button->perly_icon, XtNlabel, "", NULL);
}


static void
really_do_quit(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	if (button_list == NULL)
		exit(0);

	current_button = GetCurrentButton(w);
	if (current_button->is_executing && !current_button->has_finished) {
		interrupt(w, (XtPointer)current_button, (XtPointer)NULL);
		CleanUpOnChildEOF(current_button);
	}

	(void)XukcRemoveObjectOnList(&button_list, (XtPointer)current_button);
	update_wm_command();

	if (CurrentTailoring() == current_button);
		TailoringPopup((Widget)NULL, (XEvent *)NULL, (String *)NULL,
							(Cardinal *)NULL);
	if (button_list == NULL)
		/* all the buttons have been destroyed so kill the process */
		XtDestroyWidget(creation_button->perly);

	if (current_button->execution_directory != NULL)
		XtFree(current_button->execution_directory);
	if (current_button->filename != NULL);
		XtFree(current_button->filename);
	XtFree(current_button);
}


static void
save_perly(w, event, params, num_params)
Widget w;
XEvent *event;
String *params;
Cardinal *num_params;
{
	Save(GetCurrentButton(w), TRUE);
	update_wm_command();
	XBell(XtDisplay(w), 50);
}


static void
quit_perly(w, event, params, num_params)
Widget w;
XEvent *event;
String *params;
Cardinal *num_params;
{
	current_button = GetCurrentButton(w);
	if (current_button->is_executing && !current_button->has_finished)
		interrupt(current_button->interrupt_button,
				(XtPointer)current_button, (XtPointer)NULL);
	/* try to free all associated memory and maybe stuff in the server */
	if (*num_params == 0)
		Save(current_button, TRUE);
	else if (*num_params == 1 && strcmp(params[0], "remove") == 0)
		RemoveFile(current_button);

	if (current_button == creation_button) {
		CleanUpOnChildEOF(creation_button);
		XtUnmapWidget(creation_button->perly);
		(void)XukcRemoveObjectOnList(&button_list,
					(XtPointer)creation_button);
		XtDestroyWidget(creation_button->edit_popup);
		if (CurrentTailoring() == current_button);
			TailoringPopup((Widget)NULL, (XEvent *)NULL,
					(String *)NULL, (Cardinal *)NULL);
		if (button_list == NULL)
			/* all the buttons have been destroyed so kill the
			   process */
			XtDestroyWidget(creation_button->perly);
	} else
		XtDestroyWidget(current_button->perly);
}


static void
do_quit_button(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	Cardinal num_params = 0;

	quit_perly(w, (XEvent *)NULL, (String *)NULL, &num_params);
}


static void
button_has_moved(w, client_data, event, cont)
Widget w;
XtPointer client_data;
XEvent *event;
Boolean *cont;
{
	struct app_resources *butt = (struct app_resources *)client_data;

	if (event->type != ConfigureNotify)
		return;

	if (event->xconfigure.override_redirect)
		return;

	if (butt->edit_is_up || CurrentTailoring() == butt)
		PositionPopups(butt);
	StoreGeometry(butt);
}


static void
bind_button_interface(butt)
struct app_resources *butt;
{
	XtVaSetValues(butt->perly, XtNallowShellResize, TRUE, NULL);
	XtVaSetValues(butt->perly_button, XtNresize, TRUE, NULL);

	XtAddCallback(butt->perly_button, XtNcallback, execute,
							(XtPointer)butt);
	XtAddCallback(butt->perly, XtNdestroyCallback, really_do_quit,
							(XtPointer)butt);
	XtAddEventHandler(butt->perly, StructureNotifyMask, FALSE,
					button_has_moved, (XtPointer)butt);
	XtAddEventHandler(butt->transparent, ExposureMask, FALSE, exposure,
							(XtPointer)butt);
}


static void
create_new_button_interface(button_resources, copy_button)
struct app_resources *button_resources, *copy_button;
{
	Pixmap bitmap, left_bitmap, right_bitmap;
	XFontStruct *font;
	Pixel foreground, background, highlight, edit;

	SetInitializeAttributes(button_resources);

	if (copy_button == NULL)
		return;

	XtVaGetValues(copy_button->perly_button,
				XtNforeground, &foreground,
				XtNbackground, &background,
				XtNfont, &font,
				XtNbarColour, &edit,
				XtNhighlightColor, &highlight,
				XtNbitmap, &bitmap,
				XtNleftBitmap, &left_bitmap,
				XtNrightBitmap, &right_bitmap, NULL);

	XtVaSetValues(button_resources->perly_button,
				XtNforeground, foreground,
				XtNbackground, background,
				XtNfont, font,
				XtNbarColour, edit,
				XtNhighlightColor, highlight,
				XtNbitmap, bitmap,
				XtNleftBitmap, left_bitmap,
				XtNrightBitmap, right_bitmap, NULL);
	CopyAndStoreAttributes(copy_button, button_resources);
}


static void
create_and_initialize_button(x, y, copy_button, file)
Position x, y;
struct app_resources *copy_button;
String file;
{
	Dimension width, height;
	struct app_resources *button_resources;

	button_resources = XtNew(struct app_resources);
	bzero((char *)button_resources, sizeof(struct app_resources));

	/* initialize the server id and associated attributes to NULL */
	button_resources->this_server = 0;
	button_resources->server_id = 0;
	button_resources->num_servers = 0;
	button_resources->geometry = 0;
	button_resources->font = 0;
	button_resources->foreground = 0;
	button_resources->background = 0;
	button_resources->highlight_colour = 0;
	button_resources->edit_colour = 0;
	button_resources->bitmap = 0;
	button_resources->left_bitmap = 0;
	button_resources->right_bitmap = 0;
	button_resources->script = NULL;

	button_resources->is_executing = FALSE;
	button_resources->edit_is_up = FALSE;
	button_resources->sent_interrupt = FALSE;

	(void)XukcAddObjectToList(&button_list, (XtPointer)button_resources,
						FALSE);
	if (creation_button == NULL) {
		/* this is the master button */
		extern Widget WI_perly, WI_perly_button, WI_perly_icon;
		extern Widget WI_transparent;
		extern Widget WI_edit_popup, WI_edit_script;
		extern Widget WI_hide_edit_button;
		extern Widget WI_execute_button, WI_stderr_clear;
		extern Widget WI_stderr_text;
		extern Widget WI_stdout_clear, WI_stdout_text;
		extern Widget WI_interrupt_button;
		extern Widget WI_quit_button, WI_edit_button, WI_pause_button;

		button_resources->perly = WI_perly;
		button_resources->perly_button = WI_perly_button;
		button_resources->perly_icon = WI_perly_icon;
		button_resources->transparent = WI_transparent;
		button_resources->edit_popup = WI_edit_popup;
		button_resources->edit_script = WI_edit_script;
		button_resources->edit_button = WI_edit_button;
		button_resources->hide_edit_button = WI_hide_edit_button;
		button_resources->execute_button = WI_execute_button;
		button_resources->stderr_clear = WI_stderr_clear;
		button_resources->stderr_text = WI_stderr_text;
		button_resources->stdout_clear = WI_stdout_clear;
		button_resources->stdout_text = WI_stdout_text;
		button_resources->interrupt_button = WI_interrupt_button;
		button_resources->pause_button = WI_pause_button;
		button_resources->quit_button = WI_quit_button;
		creation_button = button_resources;
		InitializeTailoring(creation_button);
		ChangeTailoringButton(creation_button);
		BindOutputInterface(creation_button);
	} else
		CreateANewPerlyButton(creation_button->perly, button_resources);

	bind_button_interface(button_resources);

	if (file != NULL) {
		if (!Load(button_resources, file)) {
			/* loading failed so remove the button */
			if (button_resources != creation_button) {
				XtDestroyWidget(button_resources->perly);
				return;
			} else /* keep the first button even if failed load */
				SetInitializeAttributes(button_resources);
		}
		if (button_resources->script != NULL &&
		    button_resources->edit_script != NULL)
			XukcNewTextWidgetString(button_resources->edit_script,
						button_resources->script);
	} else {
		create_new_button_interface(button_resources, copy_button);

		if (x != 0 || y != 0) {
			char geometry[30];

			(void)sprintf(geometry, "+%d+%d", (int)x, (int)y);
			XtVaSetValues(button_resources->perly, XtNgeometry,
						 	geometry, NULL);
		}
	}

	XtGetApplicationResources(button_resources->perly, button_resources,
			perly_resources, XtNumber(perly_resources),
			(ArgList)NULL, 0);
	if (button_resources->save_directory != NULL)
		button_resources->save_directory =
			XtNewString(button_resources->save_directory);
	if (button_resources->execution_directory != NULL)
		button_resources->execution_directory =
			XtNewString(button_resources->execution_directory);
	else {
		char long_name[1024];

		long_name[0] = '\0';
		SetDirectory(button_resources, long_name);
	}

	InitializeFileSaving(button_resources, file);
	InitializeNewButtonForTailoring(button_resources);

	XtVaGetValues(button_resources->perly_button, XtNwidth, &width,
				XtNheight, &height, NULL);
	XtVaSetValues(button_resources->perly,
				XtNwidth, width, XtNheight, height, NULL);

	if (button_resources != creation_button)
		XtPopup(button_resources->perly, XtGrabNone);
	update_wm_command();
}


static void
create_all_buttons(argc, argv)
int *argc;
char *argv[];
{
	while (*argc > 1)
		create_and_initialize_button(0, 0, (struct app_resources *)NULL,
						argv[--(*argc)]);
}


static String
get_data_property(dpy, client_window, data_property, used)
Display *dpy;
Window client_window;
Atom data_property;
int *used;
{
#define DEFAULT_PACKET_SIZE 1024 /* number of 32-bits in data */
	Atom type;
	int nitems, format, size, left = 1, status, length, data_offset = 0;
	char *retstr = NULL;

	*used = 0;
	size = DEFAULT_PACKET_SIZE;
	while (data_property != NULL && left) {
		if (retstr != NULL)
			XtFree(retstr);

		/* retrieve the data and then delete the property which
		 * signals to the slave that the master is ready to
		 * receive more data.
		 */
		if (XGetWindowProperty(dpy, client_window, data_property,
		    data_offset, size, TRUE, XA_STRING, &type, &format, &nitems,
		    &left, &retstr) != Success)
			return NULL; /* ||| unable to retrieve data! ||| */
		if (type != XA_STRING) /* ||| hmm, donno what this data is |||*/
			return NULL;

		size += left;
	};

	*used = nitems;
	return retstr;
}


static void
new_button_signalled(w, client_data, event)
Widget w;
XtPointer client_data;
XEvent *event;
{
	String new_buttons, filename;
	int used;

	if (event->type != PropertyNotify)
		XtError("perly: new_button_signalled() called by wrong event");

	if (event->xproperty.atom != perly_new_atom)
		return;

	if (event->xproperty.state == PropertyNewValue) {
		/* the new value of the property will be a list of
		 * files or "\001"'s signify new buttons must be created.
		 */
		new_buttons = get_data_property(XtDisplay(w), XtWindow(w),
				perly_new_atom, &used);
		if (new_buttons == NULL)
			/* failed 8-{ */
			return;
		filename = new_buttons;

		while (*filename) {
			create_and_initialize_button(0, 0,
				(struct app_resources *)NULL,
				*filename == '\001' ? (String)NULL : filename);
			filename += strlen(filename) + 1;
		}
		XtFree(new_buttons);
	}
}


/* This sets up the event handler for slaves to tell the master to create
 * new buttons.  The event handler gets PropertyNotify messages saying that
 * the new_atom has been updated.  To show slaves we are the master we
 * must set the master_atom property to be a window id. that we want
 * the new_atom property to be set on.  The master_atom is a property of
 * the default root window of the display.
 */
static void
initialize_external_creation(w, master_atom, new_atom)
Widget w;
Atom master_atom, new_atom;
{
	Window win = XtWindow(w);

	perly_master_atom = master_atom;
	perly_new_atom = new_atom;

	XtAddEventHandler(w, PropertyChangeMask, FALSE, new_button_signalled,
				(XtPointer)NULL);
	XChangeProperty(XtDisplay(w), XtWindow(w),
			master_atom, XA_WINDOW, 32, PropModeReplace,
			(unsigned char *)&win, 1);
	XChangeProperty(XtDisplay(w), RootWindow(XtDisplay(w), 0),
			master_atom, XA_WINDOW, 32, PropModeReplace,
			(unsigned char *)&win, 1);
}


/*************** Module Exported Routines *******************/

/* given a widget who is a child of a perly button find the associated
 * data structure and set the "current_button" variable to it */
struct app_resources *
GetCurrentButton(w)
Widget w;
{
	struct app_resources **buttons;
	Cardinal num_buttons;

	buttons = (struct app_resources **)XukcGetListPointer(button_list,
								&num_buttons);
	if (num_buttons == 0)
		XtError("GetCurrentButton() called with no buttons created");

	w = toplevel_parent(w);

	for (; num_buttons > 0; num_buttons--)
		if (buttons[num_buttons-1]->perly == w) {
			return buttons[num_buttons-1];
		}

	XtError("Could not find data associated with widget in GetCurrentButton()");
}


void
PositionPopups(butt)
struct app_resources *butt;
{
#define BORDER 2 /* ||| should really find out from window manager! ||| */
	Position x, y, tx, ty, ex, ey;
	Dimension height, width, pt_width, pt_height, pe_width, pe_height;
	Dimension offset_x, offset_y; /* values to offset tailoring window by */
	int dw = WidthOfScreen(XtScreen(WI_tailor_popup));
	int dh = HeightOfScreen(XtScreen(WI_tailor_popup));
	Boolean popup_to_left, popup_above;

	XtVaGetValues(WI_tailor_popup, XtNx, &tx, XtNy, &ty,
					XtNwidth, &pt_width,
					XtNheight, &pt_height, NULL);
	if (OUTPUT_POPUP_CREATED(butt))
		XtVaGetValues(butt->edit_popup, XtNx, &ex, XtNy, &ey,
					XtNwidth, &pe_width,
					XtNheight, &pe_height, NULL);
	XtVaGetValues(WI_hide_tailor_button, XtNy, &y,
					XtNheight, &offset_y, NULL);
	offset_y += (Dimension)y;
#if 0
	offset_x = (Dimension)25; /* ||| change ||| */
#else
	offset_x = 0;
#endif
	XtVaGetValues(butt->perly, XtNx, &x, XtNy, &y,
				XtNheight, &height, 
				XtNwidth, &width, NULL);

	popup_to_left = popup_above = FALSE;

	/* work out where to position the popups */
	popup_to_left = (x + (Position)pt_width) > dw;
	popup_above = (y + (Position)pt_height + (Position)height) > dh;

	if (popup_to_left) {
		tx = (x - pt_width + width /*- offset_x*/);
		ex = (x - pe_width + width);
	} else {
		tx = x /*+ offset_x*/;
		ex = x;
	}

	if (popup_above) {
		ty = (y - pt_height /*- offset_y*/ - BORDER);
		ey = (y - pe_height - BORDER);
	} else {
		ty = (y + height + BORDER /*+ offset_y*/);
		ey = (y + height + BORDER);
	}

	if (OUTPUT_POPUP_CREATED(butt))
		XtVaSetValues(butt->edit_popup, XtNx, ex, XtNy, ey, NULL);

	if (CurrentTailoring() == butt)
		XtVaSetValues(WI_tailor_popup, XtNx, tx, XtNy, ty, NULL);
}


void
BindOutputInterface(butt)
struct app_resources *butt;
{
	/* ||| shouldn't have to do this!! ||| */
	XtVaSetValues(butt->stderr_text, XtNeditType, XawtextEdit, NULL);
	XtVaSetValues(butt->stdout_text, XtNeditType, XawtextEdit, NULL);
	XtVaSetValues(butt->edit_script, XtNeditType, XawtextEdit, NULL);

	if (butt->script != NULL)
		XukcNewTextWidgetString(butt->edit_script, butt->script);

	XtAddCallback(butt->quit_button, XtNcallback, do_quit_button,
							(XtPointer)NULL);
	if (butt->show_quit)
		XtManageChild(butt->quit_button);

	XtAddCallback(butt->execute_button, XtNcallback, execute,
							(XtPointer)butt);
	XtAddCallback(butt->interrupt_button, XtNcallback, interrupt,
							(XtPointer)butt);
	XtAddCallback(butt->pause_button, XtNcallback, pause, (XtPointer)butt);
	XtAddCallback(butt->edit_button, XtNcallback, edit, (XtPointer)butt);
	XtAddCallback(butt->hide_edit_button, XtNcallback, perly_popdown,
							(XtPointer)butt);
	XtAddCallback(butt->stderr_clear, XtNcallback, clear,
						(XtPointer)butt->stderr_text);
	XtAddCallback(butt->stdout_clear, XtNcallback, clear,
						(XtPointer)butt->stdout_text);
}


void
StartPerly(dpy_str, master_atom, new_atom, argc, argv)
String dpy_str;
Atom master_atom, new_atom;
int *argc;
char *argv[];
{
	/* we start up a new master perly button */
	Widget tmp;

	tmp = XukcDirtGO(dpy_str, argc, argv, options, XtNumber(options), NULL);
	XtFree(dpy_str);
	XtAppAddActions(XtWidgetToApplicationContext(tmp),
						actions, XtNumber(actions));
	XukcAddConverters();
	XukcCreateUserInterface(tmp);
	XukcOverrideConverters();
	XtSetTypeConverter(XtRString, XukcRPopupResponse,
			XukcCvtStringToNamedValue,
			popup_response_Args, XtNumber(popup_response_Args),
			XtCacheAll, (XtDestructor)NULL);
	XtSetTypeConverter(XukcRPopupResponse, XtRString,
			XukcCvtNamedValueToString,
			popup_response_Args, XtNumber(popup_response_Args),
			XtCacheAll, (XtDestructor)NULL);
	XtSetTypeConverter(XtRString, XukcRIconResponse,
			XukcCvtStringToNamedValue,
			icon_response_Args, XtNumber(icon_response_Args),
			XtCacheAll, (XtDestructor)NULL);
	XtSetTypeConverter(XukcRIconResponse, XtRString,
			XukcCvtNamedValueToString,
			icon_response_Args, XtNumber(icon_response_Args),
			XtCacheAll, (XtDestructor)NULL);

	XtAddCallback(WI_hide_tailor_button, XtNcallback, tailor_popdown,
							(XtPointer)NULL);

	if (*argc > 1) {
		create_and_initialize_button(0, 0, (struct app_resources *)NULL,
								argv[*argc - 1]);
		(*argc)--;
	} else
		create_and_initialize_button(0, 0, (struct app_resources *)NULL,
								(String)NULL);
	/* clean up on exit .. saves all buttons to files */
	(void)signal(SIGHUP, signal_save);
	(void)signal(SIGINT, signal_save);
	(void)signal(SIGTERM, signal_save);
	XSetIOErrorHandler(signal_save);

	XtRealizeWidget(tmp);
	set_wm_protocols(tmp);
	initialize_external_creation(tmp, master_atom, new_atom);
	create_all_buttons(argc, argv);
	XtAppMainLoop(XtWidgetToApplicationContext(tmp));
}
