#ifndef lint
static char rcsid[] = "$Id: startup.c,v 1.3 90/08/25 16:58:11 rlh2 Exp $";
#endif /* !lint */

/* 
 * Copyright 1990 Richard Hesketh / rlh2@ukc.ac.uk
 *                Computing Lab. University of Kent at Canterbury, UK
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of Richard Hesketh and The University of
 * Kent at Canterbury not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Richard Hesketh and The University of Kent at Canterbury make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * Richard Hesketh AND THE UNIVERSITY OF KENT AT CANTERBURY DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL Richard Hesketh OR THE
 * UNIVERSITY OF KENT AT CANTERBURY BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 *
 * Author:  Richard Hesketh / rlh2@ukc.ac.uk, 
 *                Computing Lab. University of Kent at Canterbury, UK
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/X.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>
#include <X11/Xresource.h>

#define XA_MASTER_PERLY_BUTTON_WINDOW "MASTER_PERLY_BUTTON_WINDOW"
#define XA_NEW_PERLY_BUTTON_DESCRIPTION "NEW_PERLY_BUTTON_DESCRIPTION"
#define PROG_NAME "perly"

#define FALSE 0
#define TRUE 1

extern char *malloc();
extern char *getenv();

static Atom master_perly_atom = NULL;
static Atom new_perly_atom = NULL;
static int (*old_error_handler)();

static XrmOptionDescRec options[] = {
	{ "-display", ".display", XrmoptionSepArg, (caddr_t)NULL },
};


static char *
emalloc(size)
unsigned int size;
{
	extern char *malloc();
	char *ret;

	ret = malloc(size);
	if (ret == NULL) {
		fprintf(stderr, "%s: Malloc failed in startup\n", PROG_NAME);
		exit (-1);
	}
	return (ret);
}


static void
intern_atoms(dpy)
Display *dpy;
{
	if (master_perly_atom != NULL)
		return;

	master_perly_atom = XInternAtom(dpy, XA_MASTER_PERLY_BUTTON_WINDOW,
								FALSE);
	new_perly_atom = XInternAtom(dpy, XA_NEW_PERLY_BUTTON_DESCRIPTION,
								FALSE);
}


static int
error_handler(dpy, errorp)
Display *dpy;
XErrorEvent *errorp;
{
	switch (errorp->request_code) {
		case X_GetGeometry:
			if (errorp->error_code == BadDrawable)
				break;
		case X_GetProperty:
			if (errorp->error_code == BadWindow)
				break;
		default:
			return ((*old_error_handler)(dpy, errorp));
	}
	return 0;
}
									  

static Window
get_existing_master(dpy)
Display *dpy;
{
	Atom type;
	int format;
	unsigned long nitems, left;
	Window *retwin;
	int status;
	Window win;

	intern_atoms(dpy);

	status = XGetWindowProperty(dpy, RootWindow(dpy, 0), master_perly_atom,
			0l, 4l, False, XA_WINDOW, &type, &format, &nitems,
			&left, (unsigned char **)&retwin);

	if (status == Success && type == XA_WINDOW) { /* found it! */
		win = *retwin;
		XtFree(retwin);
	} else
		win = NULL;

	if (win != NULL) { /* check that the master really still exists */
		Status ok;
		Window dummy;   /* dummy return variables */
		int x, y;
		unsigned int w, h, bw, dp;

		/* this function is only used to ask the server whether
		 * the window still exists or not. */
		old_error_handler = XSetErrorHandler(error_handler);
		ok = XGetGeometry(dpy, (Drawable)win, &dummy, &x, &y,
						&w, &h, &bw, &dp);


		if (ok)
			status = XGetWindowProperty(dpy, win,
				master_perly_atom, 0l, 4l, False, XA_WINDOW,
				&type, &format, &nitems, &left,
				(unsigned char **)&retwin);

		(void)XSetErrorHandler(old_error_handler);

		if (ok && status == Success && type == XA_WINDOW) {
			/* master is alive */
			win = *retwin;
			XtFree(retwin);
		} else
			win = NULL;
	}

	return win;
}


static void
start_a_master(dpy, argc, argv)
Display *dpy;
int *argc;
char *argv[];
{
	char *dpy_str = malloc(strlen(DisplayString(dpy)) + 1);

	(void)strcpy(dpy_str, DisplayString(dpy));
	XCloseDisplay(dpy);
	StartPerly(dpy_str, master_perly_atom, new_perly_atom, argc, argv);
}


static void
start_a_slave(dpy, argc, argv, win)
Display *dpy;
int *argc;
char *argv[];
Window win;
{
	/* To start a slave we inform the master by setting a particular
	 * property on the master's window.  This property contains
	 * the name of the description file or "\001" if no file has been
	 * given.  Because other perly's may be starting up at the same
	 * time we check to see if the property exists and if so wait for
	 * it to be deleted by the master.  Now lots of perly's may be
	 * wanting to set the property so we just append to it.
	 * We then wait for the property to be deleted, meaning that the
	 * master has taken the filename and is doing something with it.
	 * The slave start up process then dies.
	 */
	char *filename;
	char not_finished = TRUE;

	intern_atoms(dpy);

	if (*argc == 2) {
		char *cwd;

		filename = emalloc(1000);
		cwd = getenv("PWD");
		if (cwd == NULL)
			(void)strcpy(filename, argv[1]);
		else
			(void)sprintf(filename, "%s/%s", cwd, argv[1]);
	} else {
		filename = emalloc(2);
		strcpy(filename, "\001"); /* create a blank button */
	}

	XSelectInput(dpy, win, PropertyChangeMask);

	XChangeProperty(dpy, win, new_perly_atom, XA_STRING, 8, PropModeAppend,
			(unsigned char *)filename, strlen(filename) + 1);
	free(filename);

	while (not_finished) {
		XEvent event;

		XNextEvent(dpy, &event);

		if (event.xproperty.state == PropertyDelete &&
		    event.xproperty.atom == new_perly_atom)
			/* finished set up!! */
			not_finished = FALSE;
	}
}


main(argc, argv)
int argc;
char *argv[];
{
	Display *dpy;
	char resource[100];
	char *blank, *display;
	XrmDatabase db = NULL;
	XrmValue value;
	Window master_win;

	XrmParseCommand(&db, options, 1, PROG_NAME, &argc, argv);

	(void)sprintf(resource, "%s.display", PROG_NAME);
	if (XrmGetResource(db, resource, "", &blank, &value))
		display = (char *)value.addr;
	else
		display = NULL;

	if ((dpy = XOpenDisplay(display)) == NULL) {
		fprintf(stderr, "%s: Couldn't open display %s\n",
				PROG_NAME, XDisplayName(display));
		exit (-1);
	}

	master_win = get_existing_master(dpy);

	if (master_win == NULL)
		start_a_master(dpy, &argc, argv);
	else
		start_a_slave(dpy, &argc, argv, master_win);
}
