#ifndef lint
static char rcsid[] =
   "$Xukc: tailor.c,v 1.5 91/01/13 23:08:41 rlh2 Exp $";
#endif /* !lint */

/* 
 * Copyright 1990 Richard Hesketh / rlh2@ukc.ac.uk
 *                Computing Lab. University of Kent at Canterbury, UK
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of Richard Hesketh and The University of
 * Kent at Canterbury not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Richard Hesketh and The University of Kent at Canterbury make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * RICHARD HESKETH AND THE UNIVERSITY OF KENT AT CANTERBURY DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL RICHARD HESKETH OR THE
 * UNIVERSITY OF KENT AT CANTERBURY BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 *
 * Author:  Richard Hesketh / rlh2@ukc.ac.uk, 
 *                Computing Lab. University of Kent at Canterbury, UK
 */

#include <X11/Xatom.h>
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/Xukc/Toggle.h>
#include <X11/Xukc/Value.h>
#include <X11/Xukc/text.h>
#include <X11/Xukc/memory.h>
#include "perly.h"

#define PRIMARY_COLOR "PRIMARY_COLOR"
#define PRIMARY_FONT "PRIMARY_FONT"

extern struct app_resources *GetCurrentButton();
extern void PositionPopups();
extern void StoreAttribute();

extern Widget WI_tailor_icon_on_when_output, WI_tailor_icon_always_on;
extern Widget WI_tailor_icon_always_off, WI_perly_icon;
extern Widget WI_tailor_popup, WI_tailor_popup_on_output;
extern Widget WI_tailor_popup_on_completion;
extern Widget WI_tailor_popup_never, WI_beep_on_completion;
extern Widget WI_popdown_on_completion;
extern Widget WI_auto_clear, WI_pass_args, WI_split_args, WI_show_quit_button;
extern Widget WI_quit_button, WI_execution_directory, WI_font_button;
extern Widget WI_foreground_button;
extern Widget WI_background_button, WI_bar_colour_button;
extern Widget WI_highlight_colour_button;
extern Widget WI_highlight_thickness, WI_internal_height, WI_internal_width;
extern Widget WI_bitmap_button, WI_left_bitmap_button, WI_right_bitmap_button;
extern Widget WI_perly_button;

static struct app_resources *tailored_button;

void SetIconWidth();

static String existing_font = NULL;


/* The button has been edited so it has resized itself .. we therefore
 * change the size of the popup output icon to this size as well.
 * The "title" and "icon name" are changed to something that is more
 * meaningful to the user than just "perly".
 */
static void
change_button(w, client_data, event, cont)
Widget w;
XtPointer client_data;
XEvent *event;
Boolean *cont;
{
	struct app_resources *butt = (struct app_resources *)client_data;
	String name;
	char title[200];

	SetIconWidth(butt);

	/* change the icon name and title to something meaningful */
	XtVaGetValues(butt->perly_button, XtNlabel, &name, NULL);
	strncpy(title, name, 190);
	strcat(title, " (perly)");
	XtVaSetValues(butt->perly, XtNtitle, title, XtNiconName, title, NULL);
}


/* The name of the font has been edited and the perly_button's label
 * is updated to this new font.  Also the font that the edited font name
 * is displayed in is changed to this new font.
 */
static void
show_new_font(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	char check_str[201];
	String new_font_str;

	if (call_data != NULL)
		/* we have the new font name in call_data */
		new_font_str = (String)call_data;
	else
		XtVaGetValues(w, XtNlabel, &new_font_str, NULL);

	if (new_font_str == NULL || *new_font_str == '\0')
		return;

	XtVaSetValues(w, XtVaTypedArg, XtNfont, XtRString,
				new_font_str, strlen(new_font_str) + 1, NULL);
	check_str[0] = '\0';
	XtVaGetValues(w, XtVaTypedArg, XtNfont, XtRString,
				check_str, 200, NULL);
	if (existing_font != NULL && strcmp(check_str, existing_font) == 0) {
		/* The new font failed to be found .. so go back to the
		 * old one.
		 */
		XtVaSetValues(w, XtVaTypedArg, XtNfont, XtRString,
				existing_font, strlen(existing_font) + 1, NULL);
		return;
	}
	if (existing_font != NULL) {
		XtFree(existing_font);
		existing_font = XtNewString(new_font_str);
	}
	XtVaSetValues(tailored_button->perly_button, XtVaTypedArg, XtNfont, 
				XtRString,
				new_font_str, strlen(new_font_str) + 1, NULL);
	SetIconWidth(tailored_button);
	(void)StoreAttribute(tailored_button, XtNfont, new_font_str);	
}


/* The grabbing of the PRIMARY_FONT selection has either FAILed or succeeded.
 * If it has succeeded then update the perly_button's font.
 */
static void
upload_selection_value(w, client_data, selection, type, value, length, format)
Widget w;
XtPointer client_data;
Atom *selection, *type;
XtPointer value;
unsigned long *length;
int *format;
{
	if (*type == XT_CONVERT_FAIL || value == NULL)
		return;

	XukcChangeLabel(w, (String)value, FALSE);
	show_new_font(w, client_data, (XtPointer)value);
}


/* Get the intrinsics to retrieve the current_value of the PRIMARY_FONT
 * (if there is one).  The retrieval happens automatically and we are
 * informed of its result by the procedure upload_selection_value(),
 * being called.
 */
static void
get_font_selection(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	static Atom _XA_PRIMARY_FONT = NULL;

	if (_XA_PRIMARY_FONT == NULL)
		_XA_PRIMARY_FONT = XInternAtom(XtDisplay(w),
						PRIMARY_FONT, FALSE);
	XtGetSelectionValue(w, _XA_PRIMARY_FONT,
				XA_STRING, upload_selection_value,
				(XtPointer)tailored_button,
				XtLastTimestampProcessed(XtDisplay(w)));
}


static void
change_quit(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	if (!OUTPUT_POPUP_CREATED(tailored_button))
		return;

	if (tailored_button->show_quit)
		XtManageChild(tailored_button->quit_button);
	else
		XtUnmanageChild(tailored_button->quit_button);
}


static void
display_current_values()
{
#define DISPLAY_VALUE(w, r, field) { \
	str = (String)XukcGetObjectOnList(tailored_button->field, \
					  tailored_button->this_server); \
	XtVaSetValues(w, r, str, NULL); \
}
	Dimension thickness, int_width, int_height;
	String str;

	switch(tailored_button->icon_response) {
		case ICON_ON_WHEN_OUTPUT:
			XukcSetToggle(WI_tailor_icon_on_when_output, TRUE);
			break;
		case ICON_ALWAYS_ON:
			XukcSetToggle(WI_tailor_icon_always_on, TRUE);
			break;
		case ICON_ALWAYS_OFF:
			XukcSetToggle(WI_tailor_icon_always_off, TRUE);
			break;
	}

	switch(tailored_button->popup_response) {
		case POPUP_ON_OUTPUT:
			XukcSetToggle(WI_tailor_popup_on_output, FALSE);
			break;
		case POPUP_ON_COMPLETION:
			XukcSetToggle(WI_tailor_popup_on_completion, FALSE);
			break;
		case NO_POPUP:
			XukcSetToggle(WI_tailor_popup_never, FALSE);
			break;
	}

	XtVaSetValues(WI_beep_on_completion, XtNtoggleAddress,
					&tailored_button->beep, NULL);
	XtVaSetValues(WI_popdown_on_completion, XtNtoggleAddress,
				&tailored_button->popdown_on_completion, NULL);
	XtVaSetValues(WI_auto_clear, XtNtoggleAddress,
					&tailored_button->auto_clear, NULL);
	XtVaSetValues(WI_pass_args, XtNtoggleAddress,
					&tailored_button->pass_selection, NULL);
	XtVaSetValues(WI_split_args, XtNtoggleAddress,
				&tailored_button->split_selection, NULL);
	XtVaSetValues(WI_show_quit_button, XtNtoggleAddress,
					&tailored_button->show_quit, NULL);
	change_quit(WI_show_quit_button, (XtPointer)NULL,
					(XtPointer)tailored_button->show_quit);

	XtVaGetValues(tailored_button->perly_button,
		XtNhighlightThickness, &thickness,
		XtNinternalWidth, &int_width,
		XtNinternalHeight, &int_height,
		NULL);

	str = (String)XukcGetObjectOnList(tailored_button->font,
					  tailored_button->this_server);
	XtVaSetValues(WI_font_button, XtVaTypedArg, XtNfont, XtRString,
				str, strlen(str) + 1, XtNlabel, str, NULL);

	DISPLAY_VALUE(WI_foreground_button, XtNlabel, foreground);
	DISPLAY_VALUE(WI_background_button, XtNlabel, background);
	DISPLAY_VALUE(WI_bar_colour_button, XtNlabel, edit_colour);
	DISPLAY_VALUE(WI_highlight_colour_button, XtNlabel, highlight_colour);
	DISPLAY_VALUE(WI_bitmap_button, XtNlabel, bitmap);
	DISPLAY_VALUE(WI_left_bitmap_button, XtNlabel, left_bitmap);
	DISPLAY_VALUE(WI_right_bitmap_button, XtNlabel, right_bitmap);

	XtVaSetValues(WI_highlight_thickness, XtNvalue, (int)thickness, NULL);
	XtVaSetValues(WI_internal_height, XtNvalue, (int)int_height, NULL);
	XtVaSetValues(WI_internal_width, XtNvalue, (int)int_width, NULL);
	XtVaSetValues(WI_execution_directory, XtNlabel,
			tailored_button->execution_directory, NULL);
}


static void
toggle_response(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	int i = (int)call_data;
	tailored_button->popup_response = i;
}


static void
toggle_icon(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	int i = (int)call_data;
	tailored_button->icon_response = i;

	if (tailored_button->icon_response == ICON_ALWAYS_ON) {
		SetIconWidth(tailored_button);
		XtManageChild(tailored_button->perly_icon);
	} else
		XtUnmanageChild(tailored_button->perly_icon);
}


static void
change_value(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	int i = (int)call_data;

	XtVaSetValues(tailored_button->perly_button,
				(String)client_data, i, NULL);
}


/* The grabbing of the PRIMARY_COLOR selection has either FAILed or succeeded.
 * If it has succeeded then update the perly_button's colour.
 */
static void
upload_selected_colour(w, client_data, selection, type, value, length, format)
Widget w;
XtPointer client_data;
Atom *selection, *type;
XtPointer value;
unsigned long *length;
int *format;
{
	if (*type == XT_CONVERT_FAIL || value == NULL)
		return;

	XukcChangeLabel(w, (String)value, FALSE);
	XtVaSetValues(tailored_button->perly_button, XtVaTypedArg,
			(String)client_data,
			XtRString, (String)value, strlen(value) + 1, NULL);
	(void)StoreAttribute(tailored_button,
				(String)client_data, (String)value);	
}


/* Get the intrinsics to retrieve the current_value of the PRIMARY_COLOR
 * (if there is one).  The retrieval happens automatically and we are
 * informed of its result by the procedure upload_selected_colour(),
 * being called.
 */
static void
get_colour_selection(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	static Atom _XA_PRIMARY_COLOR = NULL;

	if (_XA_PRIMARY_COLOR == NULL)
		_XA_PRIMARY_COLOR = XInternAtom(XtDisplay(w),
						PRIMARY_COLOR, FALSE);

	XtGetSelectionValue(w, _XA_PRIMARY_COLOR,
				XA_STRING, upload_selected_colour,
				(XtPointer)client_data,
				XtLastTimestampProcessed(XtDisplay(w)));
}


static void
change_text_resource(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	String str = (String)call_data;

	XtVaSetValues(tailored_button->perly_button, XtVaTypedArg,
			(String)client_data,
			XtRString, str, strlen(str) + 1, NULL);
	(void)StoreAttribute(tailored_button, (String)client_data, str);	
}


static void
change_execution(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	String str = (String)call_data;

	if (tailored_button->execution_directory != NULL)
		XtFree(tailored_button->execution_directory);

	tailored_button->execution_directory = XtNewString(str);
}


/*************************** Module Exported Routines *********************/

void
InitializeTailoring(butt)
struct app_resources *butt;
{
	char font_str[500];

	XtVaGetValues(butt->perly_button, XtVaTypedArg, XtNfont, XtRString,
				font_str, 499, NULL);
	existing_font = XtNewString(font_str);

	XukcAddRadioToggle(WI_tailor_popup_on_completion,
				WI_tailor_popup_on_output);
	XukcAddRadioToggle(WI_tailor_popup_never,
				WI_tailor_popup_on_completion);
	XtAddCallback(WI_tailor_popup_on_output, XtNcallback, toggle_response,
							(XtPointer)NULL);
	XtAddCallback(WI_tailor_popup_on_completion, XtNcallback,
					toggle_response, (XtPointer)NULL);
	XtAddCallback(WI_tailor_popup_never, XtNcallback, toggle_response,
							(XtPointer)NULL);

	XukcAddRadioToggle(WI_tailor_icon_always_on,
				WI_tailor_icon_on_when_output);
	XukcAddRadioToggle(WI_tailor_icon_always_off,
				WI_tailor_icon_always_on);
	XtAddCallback(WI_tailor_icon_on_when_output, XtNcallback, toggle_icon,
							(XtPointer)NULL);
	XtAddCallback(WI_tailor_icon_always_on, XtNcallback, toggle_icon,
							(XtPointer)NULL);
	XtAddCallback(WI_tailor_icon_always_off, XtNcallback, toggle_icon,
							(XtPointer)NULL);

	XtAddCallback(WI_show_quit_button, XtNcallback, change_quit,
					(XtPointer)NULL);
	XtAddCallback(WI_execution_directory, XtNeditCallback, change_execution,
					(XtPointer)NULL);
	XtAddCallback(WI_foreground_button, XtNeditCallback,
			change_text_resource, (XtPointer)XtNforeground);
	XtAddCallback(WI_foreground_button, XtNcallback,
			get_colour_selection, (XtPointer)XtNforeground);
	XtAddCallback(WI_background_button, XtNeditCallback,
			change_text_resource, (XtPointer)XtNbackground);
	XtAddCallback(WI_background_button, XtNcallback,
			get_colour_selection, (XtPointer)XtNbackground);
	XtAddCallback(WI_bar_colour_button, XtNeditCallback,
			change_text_resource, (XtPointer)XtNbarColour);
	XtAddCallback(WI_bar_colour_button, XtNcallback,
			get_colour_selection, (XtPointer)XtNbarColour);
	XtAddCallback(WI_highlight_colour_button, XtNeditCallback,
			change_text_resource, (XtPointer)XtNhighlightColor);
	XtAddCallback(WI_highlight_colour_button, XtNcallback,
			get_colour_selection, (XtPointer)XtNhighlightColor);
	XtAddCallback(WI_font_button, XtNeditCallback,
			show_new_font, (XtPointer)FALSE);
	XtAddCallback(WI_font_button, XtNcallback,
			get_font_selection, (XtPointer)NULL);
	XtAddCallback(WI_bitmap_button, XtNeditCallback,
			change_text_resource, (XtPointer)XtNbitmap);
	XtAddCallback(WI_left_bitmap_button, XtNeditCallback,
			change_text_resource, (XtPointer)XtNleftBitmap);
	XtAddCallback(WI_right_bitmap_button, XtNeditCallback,
			change_text_resource, (XtPointer)XtNrightBitmap);
	XtAddCallback(WI_highlight_thickness, XtNvalueCallback,
			change_value, (XtPointer)XtNhighlightThickness);
	XtAddCallback(WI_internal_height, XtNvalueCallback,
			change_value, (XtPointer)XtNinternalHeight);
	XtAddCallback(WI_internal_width, XtNvalueCallback,
			change_value, (XtPointer)XtNinternalWidth);
}


void
InitializeNewButtonForTailoring(butt)
struct app_resources *butt;
{
	XtAddEventHandler(butt->perly_button, StructureNotifyMask, FALSE,
			change_button, (XtPointer)butt);
	XtSetSensitive(butt->perly_icon, FALSE);
}


struct app_resources *
CurrentTailoring()
{
	return tailored_button;
}


void
ChangeTailoringButton(butt)
struct app_resources *butt;
{
	tailored_button = butt;
	/* don't display a blank button entry */
	if (butt->font > 0)
		display_current_values();
}


void
SetIconWidth(butt)
struct app_resources *butt;
{
	Dimension width, bw;

	XtVaGetValues(butt->perly_button, XtNwidth, &width, NULL);
	XtVaGetValues(butt->perly_icon, XtNborderWidth, &bw, NULL);
	XtVaSetValues(butt->perly_icon, XtNwidth, width - bw - bw, NULL);
}


void
TailoringPopup(w, event, params, num_params)
Widget w;
XEvent *event;
String *params;
Cardinal *num_params;
{
	static Boolean tailoring_is_up = FALSE;
	struct app_resources *new_butt;


	if (w == NULL ||
	    (GetCurrentButton(w) == tailored_button && tailoring_is_up)) {
		/* called by the hide_tailor_button callback
		 * or by the button it is currently up for */
		XtPopdown(WI_tailor_popup);
		tailoring_is_up = FALSE;
	} else {
		new_butt = GetCurrentButton(w);
		ChangeTailoringButton(new_butt);
		PositionPopups(new_butt);
		XtPopup(WI_tailor_popup, XtGrabNone);
		XRaiseWindow(XtDisplay(WI_tailor_popup),
						XtWindow(WI_tailor_popup));
		tailoring_is_up = TRUE;
	}
}
