
	/**********************************************
	*  PROBLEMS:                                  *
	*  --------                                   *
	*  1. Implement "sub-forms" part of grammar.  *
	*  2. What if name or values parts of input   *
	*        are inordinately long?  It will be   *
	*	 nec. in grammar to allow for a       *
	* 	 rough length/depth to be specified   *
	*	 for the text widgets - eg. longer    *
	*        ones for addresses.  Likewise a      *
	*        cut-off length for labels, and for   *
	*        radio/toggle button forms, after     *
	*	 which they should be split over      *
	*	 two lines.			      *
	*  11. Check "illeg chars in var name" - see  *
	*        line 3 in try.                       *
	**********************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef DEBUG
#	include <X11/IntrinsicP.h>
#endif DEBUG
#include <X11/Intrinsic.h>

#if defined(XtSpecificationRelease) && XtSpecificationRelease < 5
/* BUG in Xaw/Form.h header file .. XawChain* are defined as XtChain* */
#define XawChainTop XtChainTop
#define XawChainLeft XtChainLeft
#define XawRubber XtRubber
#endif

#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Text.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Command.h>
#include <X11/Xukc/utils.h>
#include <X11/Xukc/Toggle.h>
#include <X11/Xukc/Value.h>

#include "y.tab.h"
#include "inp.h"

#ifndef PROG_NAME
#define  PROG_NAME	"xfields"
#endif /* PROG_NAME */

#define  STRLENG 200		/*  No. of chars allowed in a text widget  */
#define  HOWLONG 400		/*  Length in pixels of a text widget  */
#define  INDENT_COMMENT 25	/*  Width in pixels that a comment is indented  */

struct labels_list {
     Widget  label_widg;
     struct  labels_list  *next;
} ;

typedef struct  labels_list  LABS_LIST;

struct cop_args {
	String infile;
	String outfile;
};

typedef struct  cop_args  COPY_ARGS;

static struct app_resources {
	String	heading;
	Boolean	stay_static;
} prog_resources;

static XrmOptionDescRec options[] = {
	{ "-heading", ".heading", XrmoptionSepArg, (caddr_t)NULL },
	{ "-static", ".static", XrmoptionNoArg, (caddr_t)"true" },
};

#define offset(field) XtOffset(struct app_resources *, field)
static XtResource xfield_resources[] = {
	{"heading", "Heading", XtRString, sizeof(String),
	 offset(heading), XtRImmediate, (XtPointer)NULL},
	{"static", "Static", XtRBoolean, sizeof(Boolean),
	 offset(stay_static), XtRImmediate, (XtPointer)FALSE}
};
	 

static Boolean   shell_exists = FALSE;
static Widget    shell, form, value_w;
static XtTranslations text_trans_table = NULL;
FIL_WIDS  fwds;

extern char *emalloc();

/*------------------------------------------------------------------*/

	/*******************************
	*  Forward declare functions.  *
	*******************************/

static void makewidgets();
static void getdata();
extern char *stringsave();
int    carry_on();
char   *reason();

/*------------------------------------------------------------------*/

	/*******************************************************
	*  FUNCTION MAIN:                                      *
	*  Checks for right no. of, and existance of args.     *
	*  Calls "carry_on" to allow user to quit prematurely  *
	*  if the output file exists already.  Calls "finish_  *
	*  setup" to allow application to continue if the      *
	*  output file does not already exist.                 *
	*******************************************************/ 

void
main(argc, argv)
int       argc;
char      **argv;
{
    FILE  *fin;
    FILE  *fout;
    struct stat  inp_buf, outp_buf;
    XrmDatabase db;
    XrmValue value;
    char  *blank;
    char  resname[100];

    shell = XtAppInitialize((XtAppContext *)NULL, "XFields", 
			options, XtNumber(options), &argc, argv,
			(String *)NULL, (ArgList)NULL, 0);

    XtGetApplicationResources(shell, (XtPointer)&prog_resources,
		(XtResourceList)xfield_resources, XtNumber(xfield_resources),
		(ArgList)NULL, 0);

    /*  check for wrong number of arguments  */
    if (argc > 3)
        fprintf(stderr,
		"Usage is: %s [-static] [-heading heading] [datafile [outputfile]]\n",
		argv[0]);
    else {
        if (argc == 1) /* use stdin and stdout */
		finish_setup((String)NULL, (String)NULL);
	else if (argc == 2) { /* use input file and stdout */
		/*  check for non-existant input file  */
       
		if((fin = fopen(argv[1],"r"))==(FILE *)NULL) {
		    fprintf(stderr, "%s: cannot open %s for reading (%s)\n",
				argv[0], argv[1], reason());
		    exit (2);
		} else
		    (void)fclose(fin);
		finish_setup(argv[1], (String)NULL);
	} else if (argc == 3) { /* use input and output files */
	    /*  check for non-existant input file  */
       
	    if((fin = fopen(argv[1],"r"))==(FILE *)NULL) {
		fprintf(stderr, "%s: cannot open %s for reading (%s)\n",
				argv[0], argv[1], reason());
		exit (2);
	    } else
		(void)fclose(fin);

            /*  does outputfile already exist?  */
     
            if((fout = fopen(argv[2],"r"))!=(FILE *)NULL) {
		/*  is it the same as the input file  */

		if (stat(argv[1], &inp_buf) == -1)
		    fprintf(stderr, "%s: couldn't stat %s (%s)\n",
			    argv[0], argv[1], reason());
		if (stat(argv[2], &outp_buf) == -1)
		    fprintf(stderr, "%s: couldn't stat %s (%s)\n",
			    argv[0], argv[2], reason());

		if ((inp_buf.st_ino == outp_buf.st_ino) &&
		    (inp_buf.st_dev == outp_buf.st_dev))
		    fprintf(stderr, "%s: output file is input file\n", argv[0]);
		else
                    carry_on(argv[1], argv[2]);
	    } else /*  output file not already in existance  */
		finish_setup(argv[1], argv[2]);
	}
    }
}

/*------------------------------------------------------------------*/

	/*******************************************
	*  FUNCTION REASON:                        *
	*  Return sys_errlist[errno] if in range.  *
	*******************************************/

char *
reason()
{
    extern char *sys_errlist[];
    extern int  errno, sys_nerr;

    return ((errno > 0 && errno < sys_nerr) ?
		sys_errlist[errno] : "unknown reason");
}

/*------------------------------------------------------------------*/

	/******************************************************
	*  FUNCTION FINISH_SETUP:                             *
	*  Called by "main" and "early_exit".                 *
	*  Does the rest of the necessary                     *
	*  setting up to allow the application to continue.   *
	*  Calls "yyparse" to parse the input file, and       *
	*  "makewidgets" to create the widgets.               *
	******************************************************/

finish_setup(infile, outfile)
String infile;
String outfile;
{
    extern char   *error_string;
    extern char   *buffer;
    extern int    error_count;
    extern int    widgets_found;
    FILE  *fout;
    extern  FILE  *yyin;
            
    /*  initialize buffers, error counter  */
         
    error_string = stringsave(" ");
    buffer = stringsave(" ");

    /*  Begin lexical analysis/parsing of input file.  */
    /*  'yyin' is the input file expected by Lex.      */

    if (infile != NULL)
	yyin = fopen(infile,"r");	/*  already checked for failure  */
    yyparse();
    (void)fclose(yyin);
    if (error_count > 0)
    {
	if (error_count == 1)
            fprintf(stderr,
		"\n  %s: 1 error found - form generation suppressed\n\n",
		PROG_NAME);
	else
            fprintf(stderr,
		"\n  %s: %d errors found - form generation suppressed\n\n",
		PROG_NAME, error_count);
    
        /*  exit X if a window was created  */
    
    	if (shell_exists == TRUE)
            exit(2);
    } else {
        if (widgets_found == 'n')
        {
	    fprintf(stderr, "%s: no widgets to make\n", PROG_NAME);
    
            /*  exit X if a window was created  */
    
    	    if (shell_exists == TRUE)
                exit(2);
        } else {
            /*  call widget creation routines  */
    
	    if (outfile == NULL) /* stdout */
		makewidgets(infile, stdout);
	    else {
		if ((fout = fopen(outfile,"w")) != (FILE *)NULL)
			makewidgets(infile, fout);
		else
			fprintf(stderr, "%s: cannot open %s for writing\n",
					PROG_NAME, outfile);
	    }
        }
    }
}

/*------------------------------------------------------------------*/

	/*****************************************************
	*  FUNCTION  EARLY_EXIT:                             *
	*  Called by "carry_on".  Allows user to exit early  *
	*  and so avoid overwriting a file which already     *
	*  exists, or to continue regardless.  Calls         *
	*  "finish_setup" if user decides to continue.       *
	*****************************************************/
int
early_exit(w,client_data,call_data)
Widget   w;
XtPointer  client_data, call_data;
{
    COPY_ARGS  *cops = (COPY_ARGS *)client_data;
    if (strcmp(XtName(w), "continue") == 0)
    {
	XtUnmapWidget(shell);
	finish_setup(cops->infile, cops->outfile);
    }
    else
    {
	exit(1);
    }
}

/*------------------------------------------------------------------*/

	/********************************************************
	*  FUNCTION CARRY_ON:                                   *
	*  Called by "main".  Throws up a window which allows   *
	*  the user to choose to continue or quit in a case     *
	*  where the chosen output file already exists.  Calls  *
	*  "early_exit" to do the exiting or continuing.        *
	********************************************************/

int
carry_on(infile, outfile)
String infile, outfile;
{
    Arg  args[10];
    int  l;
    char mesg[100];
    Widget  title_form, prog_title, label, togform, cont, stop;
    Pixel   background, foreground;
    COPY_ARGS  *copies;

    if (outfile != NULL)
	sprintf(mesg, "Overwrite '%s'?", outfile);

    shell_exists = TRUE;

    form = XtCreateManagedWidget("form", formWidgetClass, shell,
			NULL, 0);

    l = 0;
    XtSetArg(args[l], XtNborderWidth, 0);				l++;
    title_form = XtCreateManagedWidget("title_form", formWidgetClass,
			form, args, l);

    l = 0;
    XtSetArg(args[l], XtNlabel, PROG_NAME);				l++;
    prog_title = XtCreateManagedWidget("prog_title", labelWidgetClass,
			title_form, args, l);
    
    l = 0;
    XtSetArg(args[l], XtNbackground, &foreground);			l++;
    XtSetArg(args[l], XtNforeground, &background);			l++;
    XtGetValues(prog_title, (ArgList)args, l);

    l = 0;
    XtSetArg(args[l], XtNbackground, background);			l++;
    XtSetArg(args[l], XtNforeground, foreground);			l++;
    XtSetValues(prog_title, (ArgList)args, l);

    l = 0;
    XtSetArg(args[l], XtNfromVert,(XtArgVal)title_form);		l++;
    XtSetArg(args[l], XtNborderWidth, 0);				l++;
    XtSetArg(args[l], XtNlabel, mesg);					l++;
    label = XtCreateManagedWidget("label", labelWidgetClass, form,
			args, l);

    l = 0;
    XtSetArg(args[l], XtNfromVert,(XtArgVal)label);			l++;
    XtSetArg(args[l], XtNborderWidth, 0);				l++;
    togform = XtCreateManagedWidget("togform", formWidgetClass,
			form, args, l);

    copies = (COPY_ARGS *) emalloc(sizeof(COPY_ARGS));
    copies->infile = infile;
    copies->outfile = outfile;

    l = 0;
    cont = XtCreateManagedWidget("continue", ukcToggleWidgetClass, togform,
			(ArgList)args, l);
    XtAddCallback(cont, XtNcallback, early_exit, (XtPointer)copies);

    l = 0;
    XtSetArg(args[l],XtNfromHoriz,(XtArgVal)cont);			l++;
    stop = XtCreateManagedWidget("quit", ukcToggleWidgetClass, togform,
			(ArgList)args, l);
    XtAddCallback(stop, XtNcallback, early_exit, (XtPointer)copies);

    XtRealizeWidget(shell);
    XtAppMainLoop(XtWidgetToApplicationContext(shell));
}

/*------------------------------------------------------------------*/

	/*********************************************************
	*  FUNCTION FILE_CHOICES:                                *
	*  Takes a pointer to the outputfile, and a pointer to   *
	*  the list of structures containing the widget info.    *
	*  (WIDG_INSTANCE *)  Writes each variable name to the   *
	*  output file.  Called by "Quit".  Calls "file_ch_list" *
	*  to write out the user's chosen value for each widget, *
	*  and itself recursively to work through the list of    *
	*  widgets.                                              *
	*********************************************************/

file_choices(fo, head)
FILE  *fo;
WIDG_INSTANCE  *head;
{
    if (head != (WIDG_INSTANCE *)NULL)
    {
	/*  Don't write comments to output file  */

        if (strcmp(head->widg_type, "comment") != 0) {
            fprintf(fo, "%s='", head->var_name);
	    file_ch_list(fo, head->val_list, 'y');
	    fprintf(fo, "'\n");
	}
	file_choices(fo, head->next);
    }
}

/*------------------------------------------------------------------*/

	/*******************************************************
	*  FUNCTION FILE_CH_LIST:                              *
	*  Takes a pointer to the output file, and a pointer   *
	*  to the widget's list of values (LIST_VALS *).       *
	*  Writes the chosen value(s) to the output file.      *
	*  Called by "file_choices", and calls itself recurs-  *
	*  ively to work through the list of widget values.    *
	*******************************************************/

file_ch_list(fo, start, first)
FILE  *fo;
LIST_VALS  *start;
int  first;
{
    if (start != (LIST_VALS *)NULL)
    {
	if ((start->chosen == TRUE) && (strcmp(start->val, "\0") != 0))
	{
	    if (first == 'y')
	    {
	        fprintf(fo, "%s", start->val);
		first = 'n';
	    }
	    else
	        fprintf(fo, " %s", start->val);
	}
	file_ch_list(fo, start->next, first);
    }
}

/*------------------------------------------------------------------*/

void
Cancel(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	exit(1);
}

	/********************************************************
	*  FUNCTION QUIT:                                       *
	*  Takes the current widget, client_data and call_data  *
	*  as args.  Exits the program, copying user data to    *
	*  the output file.  Called when the "quit" button is   *
	*  pressed (in "makewidgets").  Calls "file_choices"    *
        *  to do the writing.                                   *
	********************************************************/

void
Quit(w, client_data, call_data)
Widget   w;
XtPointer  client_data;
XtPointer  call_data;
{
    FIL_WIDS  *copy_struc;
    static Arg   args[10];
    static int   l=0;
    static char  *message;
    int    i;

    copy_struc = (FIL_WIDS *)client_data;

    XtSetArg(args[l],XtNlabel,(XtArgVal)&message);			l++;

    XtGetValues(w,args,l);
    file_choices(copy_struc->outp, copy_struc->widglist);

    if (prog_resources.stay_static)
	    (void)fflush(copy_struc->outp);
    else {
	    (void)fclose(copy_struc->outp);
	    exit(0);
    }
}

/*------------------------------------------------------------------*/

	/*****************************************************
	*  FUNCTION CREATE_TEXTWS:                           *
	*  Takes a pointer to the list of a widget's values  *
	*  (LIST_VALS *), the form which will surround that  *
	*  widget (Widget), and the label containing the     *
	*  widget's question (Widget).  Creates the text     *
        *  widgets.  Called by "create_labs".                *
	*****************************************************/

void
create_textws(start,inner_form,label)
LIST_VALS  **start;
Widget  inner_form, label;
{
    Arg     args[15];
    int     l;
    char    *strng;


    strng = (char *)XtMalloc(STRLENG);
    if (*start != (LIST_VALS *)NULL) {

	/*  Copy string in "start" into "strng", free original space  */

        strncpy(strng, (*start)->val,STRLENG);
        XtFree((*start)->val);
    } else {
	*start = (LIST_VALS *)XtNew(LIST_VALS);
        *strng = '\0';
	(*start)->next = NULL;
	(*start)->chosen = TRUE;
    }

    /*  Re-read "strng" into "start", but now with length "STRLENG"  */
    /*  - thus allowing for new answer to be longer than the default  */

    (*start)->val = strng;
    l = 0;
    XtSetArg(args[l],XtNfromHoriz,(XtArgVal)label);			l++;
    XtSetArg(args[l],XtNeditType, XawtextEdit);				l++;
    XtSetArg(args[l],XtNwidth,HOWLONG);					l++;
    XtSetArg(args[l],XtNstring,strng);					l++;
    XtSetArg(args[l],XtNlength,STRLENG);				l++;
    XtSetArg(args[l],XtNuseStringInPlace,TRUE);				l++;
    XtSetArg(args[l],XtNleft, XawChainLeft);                    l++;
    XtSetArg(args[l],XtNtop, XawChainTop);			l++;
    XtSetArg(args[l],XtNbottom, XawChainTop);			l++;
    XtSetArg(args[l],XtNdisplayCaret, FALSE);			l++;
    value_w = XtCreateManagedWidget("text", asciiTextWidgetClass,
				inner_form, args, l);
    XtVaSetValues(inner_form, XtNright, XawRubber, NULL);

    /*  Turn off return key for text widgets  */

    XtOverrideTranslations(value_w, text_trans_table);
}

/*------------------------------------------------------------------*/

	/********************************************************
	*  FUNCTION CREATE_TOGGLEWS:                            *
	*  Takes a pointer to the list of a widget's values     *
	*  (LIST_VALS *), the form which will surround that     *
	*  widget (Widget), and the label containing the        *
	*  widget's question (Widget).  Creates the toggle      *
        *  widgets.  Called by "create_labs".			*
	********************************************************/

void
create_togglews(start,inner_form,label)
LIST_VALS  **start;
Widget  inner_form, label;
{
    Arg     args[10];
    int     l;
    char    *strng;
    LIST_VALS  *p;


    value_w=label;
    for (p= *start; p!=(LIST_VALS *)NULL; p=p->next)
    {
	/*  Copy string in "start" into "strng", free original space  */

        strng = (char *)XtMalloc(STRLENG);
        strncpy(strng, p->val,STRLENG);
        XtFree(p->val);

        /*  Re-read "strng" into "start", but now with length "STRLENG"  */
        /*  - thus allowing for new answer to be longer than the default  */

        p->val = strng;
        l = 0;
        XtSetArg(args[l],XtNfromHoriz,(XtArgVal)value_w);		l++;
	XtSetArg(args[l],XtNtoggleAddress, &(p->chosen));		l++;
        value_w = XtCreateManagedWidget(strng, ukcToggleWidgetClass,
                                            inner_form, (ArgList)args, l);
    }
}

/*------------------------------------------------------------------*/

	/********************************************************
	*  FUNCTION CREATE_RADIOWS:                             *
	*  Takes a pointer to the list of a widget's values     *
	*  (LIST_VALS *), the form which will surround that     *
	*  widget (Widget), and the label containing the        *
	*  widget's question (Widget).  Creates the radio       * 
        *  widgets.  Called by "create_labs".			*
	********************************************************/

void
create_radiows(start,inner_form,label)
LIST_VALS  **start;
Widget  inner_form, label;
{
    Arg     args[10];
    int     l;
    char    *strng;
    LIST_VALS  *p;
    Widget  prev_toggle = NULL;


    value_w=label;
    for (p= *start; p!=(LIST_VALS *)NULL; p=p->next)
    {

	/*  Copy string in "start" into "strng", free original space  */

        strng = (char *)XtMalloc(STRLENG);
        strncpy(strng, p->val,STRLENG);
        XtFree(p->val);

        /*  Re-read "strng" into "start", but now with length "STRLENG"  */
        /*  - thus allowing for new answer to be longer than the default  */

        p->val = strng;
        l = 0;
        XtSetArg(args[l],XtNfromHoriz,(XtArgVal)value_w);		l++;
	XtSetArg(args[l],XtNtoggleAddress, &(p->chosen));		l++;
        XtSetArg(args[l],XtNradioToggle, prev_toggle);			l++;
        prev_toggle = value_w = XtCreateManagedWidget(strng, ukcToggleWidgetClass,
                                            inner_form, (ArgList)args, l);
    }
}

/*------------------------------------------------------------------*/

	/********************************************************
	*  FUNCTION CREATE_VALUEWS:                             *
	*  Takes a pointer to the list of a widget's values     *
	*  (LIST_VALS *), the form which will surround that     *
	*  widget (Widget), and the label containing the        *
	*  widget's question (Widget).  Creates a value         * 
        *  widget.   Called by "create_labs".			*
	********************************************************/

static void
value_changed(w, client_data, call_data)
Widget w;
XtPointer client_data, call_data;
{
	String str = (String)client_data;
	int val = (int)call_data;

	(void)sprintf(str, "%d", val);
}

void
create_valuews(start,inner_form,label)
LIST_VALS  **start;
Widget  inner_form, label;
{
    Arg     args[10];
    int     defv;
    LIST_VALS  *p;
    Widget  prev_toggle = NULL;


    p = *start;
    defv = atoi(p->next->next->val),

    value_w = XtVaCreateManagedWidget("valueEntry", ukcValueWidgetClass,
					inner_form,
					XtNfromHoriz, label,
					XtNmin, atoi(p->val),
					XtNmax, atoi(p->next->val),
					XtNvalue, defv,
					NULL);
    XtFree(p->val);
    XtFree(p->next->val);
    XtFree(p->next->next->val);
    XtFree(p->next->next);
    XtFree(p->next);
    p->next = NULL; /* chop the end off */
    p->val = XtMalloc(20); /* big enough to hold a large number! */
    (void)sprintf(p->val, "%d", defv);
    p->chosen = TRUE;
    XtAddCallback(value_w, XtNvalueCallback, value_changed,
						(XtPointer)p->val);
}


/*------------------------------------------------------------------*/

	/*********************************************************
	*  FUNCTION CREATE_LABS:                                 *
	*  Takes a pointer to the list of structures containing  *
	*  info. about each widget (WIDG_INSTANCE *), the outer  *
	*  form (Widget), the inner form which will surround the *
	*  current ques/values pair (Widget), the number of the  *
	*  curent widget, and the final inner form (Widget) for  *
	*  chaining the quit button from.  Creates the inner     *
    	*  forms.  Creates the label widgets.  Calls the         *
	*  "create_*" functions to build the associated value    *
	*  widgets.  Call itself recusively to work through the  *
	*  list of structures. Called by "makewidgets".          *
	*********************************************************/

LABS_LIST *
create_labs(head, form, inner_form, last_inner, recent_lab)
Widget    form, inner_form, *last_inner, recent_lab;
WIDG_INSTANCE   *head;
{
    int   l;
    Arg   args[15];
    Widget  label;
    LABS_LIST  *lab_li;

    if (head != (WIDG_INSTANCE *)NULL)
    {
	/*  Create the borderless forms to contain the question/val(list) pairs  */

        if (strcmp(head->widg_type, "comment") == 0) {
	    /*  Save previous question widgets for max width calculation later  */

    	    lab_li = (LABS_LIST *) emalloc(sizeof(LABS_LIST));
	    lab_li->label_widg = recent_lab;
	    lab_li->next = NULL;

	    /*  Don't create new inner_form for comment  */

            l = 0;
	    XtSetArg(args[l],XtNhorizDistance,INDENT_COMMENT);		l++;
            XtSetArg(args[l],XtNfromVert,(XtArgVal)recent_lab);		l++;
            XtSetArg(args[l],XtNjustify,XtJustifyLeft);			l++;
            XtSetArg(args[l],XtNlabel,(XtArgVal)head->question);		l++;
            XtSetArg(args[l],XtNborderWidth, 0);			l++;
            XtSetArg(args[l],XtNresize, FALSE);				l++;
	    XtSetArg(args[l],XtNright, XawChainLeft);			l++;
            recent_lab = label = XtCreateManagedWidget("field", labelWidgetClass,
                                           inner_form, args, l);
	} else {
	    l = 0;
	    XtSetArg(args[l],XtNfromVert,(XtArgVal)inner_form);		l++;
            XtSetArg(args[l],XtNborderWidth, 0);			l++;
            XtSetArg(args[l],XtNright, XawChainLeft);                    l++;
            XtSetArg(args[l],XtNleft, XawChainLeft);                    l++;
	    XtSetArg(args[l],XtNtop, XawChainTop);			l++;
	    XtSetArg(args[l],XtNbottom, XawChainTop);			l++;
    
	    inner_form = XtCreateManagedWidget("inner_form", formWidgetClass,
	    			        form, args, l);
    
	    /*  Create the labels containing the questions  */
    
            l = 0;
            XtSetArg(args[l],XtNjustify,XtJustifyLeft);			l++;
            XtSetArg(args[l],XtNlabel,(XtArgVal)head->question);		l++;
            XtSetArg(args[l],XtNborderWidth, 0);			l++;
            XtSetArg(args[l],XtNresize, FALSE);				l++;
            XtSetArg(args[l],XtNright, XawChainLeft);                    l++;
            XtSetArg(args[l],XtNleft, XawChainLeft);                    l++;
	    XtSetArg(args[l],XtNtop, XawChainTop);			l++;
	    XtSetArg(args[l],XtNbottom, XawChainTop);			l++;
            recent_lab = label = XtCreateManagedWidget("field", labelWidgetClass,
                                           inner_form, args, l);

	    /*  Save question widgets for max width calculation later  */

    	    lab_li = (LABS_LIST *) emalloc(sizeof(LABS_LIST));
	    lab_li->label_widg = label;
	    lab_li->next = NULL;
	}

	/*  Call functions to create the associated value(list)s  */

        if (strcmp(head->widg_type, "text") == 0)
            create_textws(&(head->val_list),inner_form,label);
        else if (strcmp(head->widg_type, "toggle") == 0)
            create_togglews(&(head->val_list),inner_form,label);
        else if (strcmp(head->widg_type, "radio") == 0)
            create_radiows(&(head->val_list),inner_form,label);
        else if (strcmp(head->widg_type, "value") == 0)
            create_valuews(&(head->val_list),inner_form,label);
	else  	/*  comment found, no further action nec.  */
	    { ; }

        if (head->next != (WIDG_INSTANCE *)NULL)
	    lab_li->next = create_labs(head->next, form, inner_form, last_inner, recent_lab);
        else
	{
	    /*  Save last instance of inner form to chain "quit" button from  */

	    *last_inner = inner_form;
        }
    }
    return(lab_li);
}

/*------------------------------------------------------------------*/

	/********************************************************
	*  FUNCTION LABEL_WIDTHS:                               *
	*  Takes the number of input lines.                     *
	*  Sets the widths of all the labels to the width of    *
	*  the widest label.                                    *
	*  Called by "makewidgets".                             *
	********************************************************/

static void
label_widths(ylnno, lab)

int  ylnno;
LABS_LIST  *lab;

{
    Dimension  temp, widest=0;
    Arg        args[10];
    LABS_LIST  *p;

    for (p = lab; p!=(LABS_LIST *)NULL; p=p->next)
    {
        XtSetArg(args[0],XtNwidth,&temp);
	XtGetValues(p->label_widg,(ArgList)args,1);
	if (temp>widest)
	    widest=temp;
    }
    XtSetArg(args[0],XtNwidth,widest);
    for (p = lab; p!=(LABS_LIST *)NULL; p=p->next) 
 	XtSetValues(p->label_widg,(ArgList)args,1);
}

/*------------------------------------------------------------------*/

	/********************************************************
	*  FUNCTION LABEL_HEIGHTS:                              *
	*  Takes the number of input lines.                     *
	*  Get heights for labels - these need to be recalcul-  *
	*  ated since the labels have border widths of '0'.     *
	*  Called by "makewidgets".                             *
	*  THIS IS NOT NEC. AT THE MO., BUT WILL BECOME SO      *
	*      WHEN ALLOWING FOR MULTI-LINE WIDGETS.            *
	********************************************************/

static void
label_heights(ylnno, lab)

int  ylnno;
LABS_LIST  *lab;

{
    Dimension  high, bord;
    int        l;
    Arg        args[10];
    LABS_LIST  *p;

    l = 0;
    XtSetArg(args[l],XtNheight,&high);					l++; 
    XtSetArg(args[l],XtNborderWidth,&bord);				l++; 
    XtGetValues(value_w,(ArgList)args,2); 	/*  Info from last widget  */
						/*  created - this will    */
						/*  also have to change.   */
    XtSetArg(args[0],XtNheight,high+(bord*2)); 
    for (p = lab; p!=(LABS_LIST *)NULL; p=p->next) 
        XtSetValues(p->label_widg,(ArgList)args,1); 
}

/*------------------------------------------------------------------*/

	/*******************************************************
	*  FUNCTION MAKEWIDGETS:                               *
	*  Takes the output file.  Creates                     *
	*  the shell widgets and outer form widget.  Creates   *
	*  the title label.  Calls create_labs to create       *
	*  inner forms and labels.  Creates exit button which  *
	*  calls "Quit" to write out new values to output      *
	*  file.  Calls"label_heights" to adjust the heights   *
	*  of the widget-questions.                            *
	*******************************************************/

static void
makewidgets(infile, out_f)
String infile;
FILE *out_f;
{
    Widget  new_shell, new_form, title, label, command, last_inner, rec_lab;
    Pixel   background, foreground;
    Dimension height, width;
    Arg     args[15];
    int     l, r;
    extern int  yylineno;
    extern  WIDG_INSTANCE *wid_ins;
    LABS_LIST  *labl;

    /*  Create the shell and outer form widget  */

    if (shell_exists == TRUE)
	new_shell = XtCreatePopupShell("new_shell", applicationShellWidgetClass,
			    	shell, (ArgList)args, 0);
    else
        new_shell = shell;

    text_trans_table = XtParseTranslationTable(
			"<Key>Return:	redraw-display()\n\
			 <EnterWindow>: display-caret(on)\n\
			 <LeaveWindow>: display-caret(off)");

    new_form = XtCreateManagedWidget("new_form", formWidgetClass, new_shell,
                                       NULL, 0);

    /*  Create title label  */
    
    if (infile != NULL || prog_resources.heading != NULL) {
	    l = 0;
	    XtSetArg(args[l], XtNlabel, prog_resources.heading != NULL
				? prog_resources.heading : infile);
	    l++;
            XtSetArg(args[l],XtNbottom, XawChainTop);                   l++;
            XtSetArg(args[l],XtNleft, XawChainLeft);                   l++;
            XtSetArg(args[l],XtNtop, XawChainTop);                      l++;
            XtSetArg(args[l],XtNhorizDistance, 0);                      l++;
            XtSetArg(args[l],XtNvertDistance, 0);                      l++;
            XtSetArg(args[l],XtNresizable, TRUE);                      l++;
	    label = title = XtCreateManagedWidget("title", labelWidgetClass,
						new_form, args, l);
	    l = 0;
	    XtSetArg(args[l], XtNbackground, &foreground);	l++;
	    XtSetArg(args[l], XtNforeground, &background);	l++;
	    XtGetValues(label, (ArgList)args, l);

	    l = 0;
	    XtSetArg(args[l], XtNbackground, background);	l++;
	    XtSetArg(args[l], XtNforeground, foreground);	l++;
	    XtSetValues(label, (ArgList)args, l);
    } else
	    label = title = NULL;

    /*  Call function to create the labels  */

    labl = create_labs(wid_ins, new_form, label, &last_inner, rec_lab);
    label = last_inner;

    /*  Create the "Okay" and "Cancel" buttons for exiting  */

    fwds.widglist = wid_ins;
    fwds.outp = out_f;

    /* output is only generated is "Okay" is pressed */
    l = 0;
    XtSetArg(args[l],XtNfromVert,(XtArgVal)label);	l++;
    XtSetArg(args[l],XtNlabel,(XtArgVal)"Okay");	l++;
    XtSetArg(args[l],XtNbottom, XawChainTop);                   l++;
    XtSetArg(args[l],XtNright, XawChainLeft);                   l++;
    XtSetArg(args[l],XtNtop, XawChainTop);                      l++;
    XtSetArg(args[l],XtNleft, XawChainLeft);                    l++;
    command=XtCreateManagedWidget("okay",commandWidgetClass,
                                   new_form, args, l);
    XtAddCallback(command,XtNcallback,Quit,(XtPointer)&fwds);

    l = 0;
    XtSetArg(args[l],XtNfromVert,(XtArgVal)label);	l++;
    XtSetArg(args[l],XtNfromHoriz,(XtArgVal)command);	l++;
    XtSetArg(args[l],XtNlabel,(XtArgVal)"Cancel");	l++;
    XtSetArg(args[l],XtNbottom, XawChainTop);                   l++;
    XtSetArg(args[l],XtNright, XawChainLeft);                   l++;
    XtSetArg(args[l],XtNtop, XawChainTop);                      l++;
    XtSetArg(args[l],XtNleft, XawChainLeft);                    l++;
    command=XtCreateManagedWidget("cancel",commandWidgetClass,
                                   new_form, args, l);
    XtAddCallback(command,XtNcallback,Cancel,(XtPointer)NULL);

    /*  Find width of widest label and set rest to this  */

    label_widths(yylineno, labl);

    /*  Set the height of the labels  */

    label_heights(yylineno, labl);


    /*  Realize the whole thing  */

    if (shell_exists == TRUE) {
	XtPopup(new_shell, XtGrabNone);

	/* make shell only resizable in the X direction */
	XtVaGetValues(new_shell, XtNheight, &height, XtNwidth, &width, NULL);
	XtVaSetValues(new_shell, XtNmaxHeight, height,
			XtNminHeight, height, XtNminWidth, width, NULL);
	if (title != NULL)
		XtVaSetValues(title, XtNwidth, width, NULL);
    } else {
        XtRealizeWidget(new_shell);

	/* make shell only resizable in the X direction */
	XtVaGetValues(new_shell, XtNheight, &height, XtNwidth, &width, NULL);
	XtVaSetValues(new_shell, XtNmaxHeight, height,
			XtNminHeight, height, XtNminWidth, width, NULL);
	if (title != NULL)
		XtVaSetValues(title, XtNwidth, width, NULL);
        XtAppMainLoop(XtWidgetToApplicationContext(new_shell));
    }
}
