
			/***********************
			*  YACC DECLARATIONS.  *
			***********************/
%{
#include <stdio.h>
#include "y.tab.h"
#include "inp.h"

#define  MARGIN  5		/*  Margin to left of error messages  */

static char  *err_mesgs[] = {

  /* 00 */  "Malformed line\n",
  /* 01 */  "Missing value after '|'\n",
  /* 02 */  "Missing value after '@'\n",
  /* 03 */  "Missing value after '#'\n",
  /* 04 */  "Mixing of operators illegal\n",
  /* 05 */  "Missing question, '~' and var. name\n",
  /* 06 */  "Missing question, '~', var. name, '=' and value\n",
  /* 07 */  "Illegal character in value name\n",
  /* 08 */  "Missing value before '|'\n",
  /* 09 */  "Missing value before '@'\n",
  /* 10 */  "'#'s not permitted here\n",
  /* 11 */  "Variable name not permitted to be a default value\n",
  /* 12 */  "Radio buttons may have only one default\n",
  /* 13 */  "Strings not permitted to contain '#'s\n",
  /* 14 */  "Questions/comments not permitted to be default values\n",
  /* 15 */  "Illegal character(s) in variable name\n",
  /* 16 */  "Comments belong to the widget above them and may not be the first input\n",
  /* 17 */  "Missing question before '~'\n",
  /* 18 */  "'~' and '=' should be inverted\n",
  /* 19 */  "Missing variable name\n",
  /* 20 */  "Missing variable name, '~' and '=' inverted\n",
  /* 21 */  "Missing variable name and '='\n",
  /* 22 */  "Missing question and '~'\n",
  /* 23 */  "Missing value before '%'\n",
  /* 24 */  "Missing value after '%'\n",

};

int          widgets_found = 'n';
char         *error_string;
static int   error_code;
extern int   error_count;
extern int   yylineno;
extern int   space_created;
extern int   num_of_hashes;
extern int   where_tilde;
extern int   where_equals;
extern char  *buffer;
extern WIDG_INSTANCE  *wid_ins;

WIDG_INSTANCE  *found_widget();
extern char *realloc();
extern char *stringsave();

%}

%token   STRING  VARNAME  HASHSTRING  BADSTRING  NEWLINE
%token   EQUALS  AND_OR  EXC_OR  HASH  TILDE  PERCENT

%%
				/****************
				*  YACC RULES.  *
				****************/


file	:	/* empty  */
	|	line  file
	;

line	:	comment
	|	text
	|	value
	|	toggle
	|	radio
	|	bd_strt
	|	nostrng
	|	no_widg
	|	error NEWLINE
			{ error_found(0, MALFORMED_LINE);
			  found_newline(); }
	;

str_typ	:	VARNAME
	|	HASHSTRING
	|	STRING
	|	BADSTRING
	;

val_str	:	VARNAME
	|	HASHSTRING
	|	STRING
	|	BADSTRING
	|	TILDE
			{ error_found(where_tilde, ILLEG_CHAR_IN_VAL); }
	|	EQUALS
			{ error_found(where_equals, ILLEG_CHAR_IN_VAL); }
	|       VARNAME val_str
	|       HASHSTRING val_str
	|       STRING val_str
	|       BADSTRING val_str
	|       ill_tld val_str
	|       ill_equ val_str
	;

start	:	str_typ TILDE str_typ EQUALS
	;

bd_strt	:	str_typ EQUALS str_typ TILDE error NEWLINE
			{ error_found(where_tilde, INVERT_TILDE_EQUALS);
			  found_newline(); }
	|	str_typ TILDE EQUALS error NEWLINE
			{ error_found(where_tilde+1, MISS_VAR_NAME);
			  found_newline(); }
	|	str_typ EQUALS TILDE error NEWLINE
			{ error_found(where_tilde, MISS_VN_INV_TE);
			  found_newline(); }
	|	str_typ TILDE error NEWLINE
			{ error_found(where_tilde+1, MISS_VN_AND_EQ);
			  found_newline(); }
	|	str_typ EQUALS error NEWLINE
			{ error_found(0, MISS_QU_AND_TI);
			  found_newline(); }
	|	TILDE error NEWLINE
			{ error_found(0, MISS_QUESTION);
			  found_newline(); }
	|	EQUALS error NEWLINE
			{ error_found(0, NOT_START_EQUALS);
			  found_newline(); }
	;


comment	:	str_typ NEWLINE
			{ if (error_count == 0)
			  {
			      if (widgets_found == 'y')
			          wid_ins = found_widget(wid_ins, "comment", 'n');
			      else
			      {
			          error_found(0, COMM_NOT_FIRST);
				  found_newline();
			      }
			  }
			  else
			  {
			      if (strlen(error_string) > 1)
                                  found_newline();
			      else
			      {
			          free(buffer);
				  buffer = NULL;
				  num_of_hashes = 0;
				  space_created = 'n';
			      }
			  }
			}
	;

text	:	start txt_end
	;

txt_end	:	NEWLINE
			{ if (error_count == 0)
			      wid_ins = found_widget(wid_ins, "text", 'n');
			  else
			  {
			      if (strlen(error_string) > 1)
			          found_newline();
			      else
			      {
			          free(buffer);
				  buffer = NULL;
				  num_of_hashes = 0;
				  space_created = 'n';
			      }
			  }
			}
	|	val_str NEWLINE
			{ if (error_count == 0)
			      wid_ins = found_widget(wid_ins, "text", 'y');
			  else
			  {
			      if (strlen(error_string) > 1)
			          found_newline();
			      else
			      {
			          free(buffer);
				  buffer = NULL;
				  num_of_hashes = 0;
				  space_created = 'n';
			      }
			  }
			}
	;

value	:	start val_end
	;

val_end :	val_str PERCENT val_str NEWLINE
			{ if (error_count == 0)
			      wid_ins = found_widget(wid_ins, "value", 'n');
			  else
			  {
			      if (strlen(error_string) > 1)
			          found_newline();
			      else
			      {
			          free(buffer);
				  buffer = NULL;
				  num_of_hashes = 0;
				  space_created = 'n';
			      }
			  }
			}
	|	val_str PERCENT val_end
	|	val_str PERCENT NEWLINE
			{ error_found(strlen(buffer), MISS_AFTER_PERCENT);
			  found_newline(); }
	|	val_str PERCENT bad_hsh NEWLINE
			{ error_found(strlen(buffer), MISS_AFTER_HASH);
			  found_newline(); }
	|	mixval badline
	;

toggle	:	start tog_end
	;

tog_end	:	val_str AND_OR val_str NEWLINE
			{ if (error_count == 0)
			      wid_ins = found_widget(wid_ins, "toggle", 'n');
			  else
			  {
			      if (strlen(error_string) > 1)
			          found_newline();
			      else
			      {
			          free(buffer);
				  buffer = NULL;
				  num_of_hashes = 0;
				  space_created = 'n';
			      }
			  }
			}
	|	val_str AND_OR tog_end
	|	val_str AND_OR NEWLINE
			{ error_found(strlen(buffer), MISS_AFTER_ANDOR);
			  found_newline(); }
	|	val_str AND_OR bad_hsh NEWLINE
			{ error_found(strlen(buffer), MISS_AFTER_HASH);
			  found_newline(); }
	|	mixtog badline
	;

radio	:	start rad_end
	;

rad_end	:	val_str EXC_OR val_str NEWLINE
			{ if (error_count == 0)
			  {
			      if (num_of_hashes == 0)
			          wid_ins = found_widget(wid_ins, "radio", 'y');
			      else if (num_of_hashes == 1)
			          wid_ins = found_widget(wid_ins, "radio", 'n');
			  }
			  else
			  {
			      if (strlen(error_string) > 1)
			          found_newline();

	    		      /*  If radio widget, cannot be more than 1 default  */
	    
	                      else
			      {
				  if (num_of_hashes > 1)
			          {
			              error_found(0, TOO_MANY_DEFS);
			              found_newline();
			          }
				  else
				  {
				      free(buffer);
				      buffer = NULL;
				      num_of_hashes = 0;
				      space_created = 'n';
				  }
			      }
			  }
			}		
	|	val_str EXC_OR rad_end
	|	val_str EXC_OR NEWLINE
			{ error_found(strlen(buffer), MISS_AFTER_EXCOR);
			  found_newline(); }
	|	val_str EXC_OR bad_hsh NEWLINE
			{ error_found(strlen(buffer), MISS_AFTER_HASH);
			  found_newline(); }
	|	mixrad badline
	;


no_widg	:	start wro_and tog_end
	|	start wro_and rad_end
	|	start wro_and badline
	|	start wro_exc tog_end
	|	start wro_exc rad_end
	|	start wro_exc badline
	;

nostrng	:       bad_and badline
	|       bad_exc badline
	|	bad_hsh badline
	;

mixval	:	val_str PERCENT val_str AND_OR
			{ error_found(strlen(buffer), MIXED_OPS); }
	|	val_str PERCENT val_str EXC_OR
			{ error_found(strlen(buffer), MIXED_OPS); }
	;

mixtog	:	val_str AND_OR val_str EXC_OR
			{ error_found(strlen(buffer), MIXED_OPS); }
	;

mixrad	:	val_str EXC_OR val_str AND_OR
			{ error_found(strlen(buffer), MIXED_OPS); }
	;

ill_tld :	TILDE
			{ error_found(where_tilde, ILLEG_CHAR_IN_VAL); }
	;

ill_equ :	EQUALS
			{ error_found(where_equals, ILLEG_CHAR_IN_VAL); }
	;

wro_and	:	AND_OR
			{ error_found(strlen(buffer), MISS_BEF_ANDOR); }
	;

wro_exc	:	EXC_OR
			{ error_found(strlen(buffer), MISS_BEF_EXCOR); }
	;

bad_and	:	AND_OR
			{ error_found(0, NOT_START_ANDOR); }
	;

bad_exc	:	EXC_OR
			{ error_found(0, NOT_START_EXCOR); }
	;

bad_hsh	:	HASH
			{ error_found(strlen(buffer), NOT_START_HASH); }

badline :	error NEWLINE
			{ found_newline(); }
	;

%%
				/****************
				*  PROGRAMS...  *
				****************/

/*---------------------------------------------------------------------*/

	/**************************************
	*  FUNCTION YYERROR:                  *
	*  Standard error message from Yacc.  *
	**************************************/

yyerror(mesg)
char  *mesg;
{
    fprintf("Parser error:%s\n", mesg);
}

/*---------------------------------------------------------------------*/

	/************************************************************
	*  FUNCTION FOUND_NEWLINE:                                  *
	*  Print the current input line and its error messages (if  *
	*  the line is in error).  Reset the current line buffer    *
	*  and errors buffer to empty.  Called from within the      *
	*  Yacc rules whenever an entire line has been parsed.      *
	************************************************************/

found_newline()
{
    fprintf(stderr, "%3d:  %s\n", (yylineno-1), buffer);
    fprintf(stderr, error_string);
    free(error_string);
    error_string = stringsave(" ");
    free(buffer);
    buffer = NULL;
    num_of_hashes = 0;
    space_created = 'n';
}

/*---------------------------------------------------------------------*/

	/**********************************************************
	*  FUNCTION FOUND_WIDGET:                                 *
        *  Add a widget's type to info in list of structures      *
	*  Takes a pointer to a structure containing widget       *
	*  info. (WIDG_INSTANCE *), and a string containing the   *
	*  type of the widget (pointer to char).  Called from     *
	*  within the Yacc rules whenever a valid widget has      *
	*  been found in input, and providing there have been no  *
	*  previous errors.  Calls itself recursively to work     *
	*  through the list of structures.                        *
	**********************************************************/

WIDG_INSTANCE *
found_widget(head,widt,select_button)
WIDG_INSTANCE *head;
char  *widt;
char  select_button;
{
    if (head->next == (WIDG_INSTANCE *)NULL)
    {
        head->widg_type = (char *)emalloc(strlen(widt)+1);
        strcpy(head->widg_type, widt);
        free(buffer);
	buffer = NULL;
        num_of_hashes = 0;
	space_created = 'n';
	if (strcmp(widt, "comment") != 0)

            /*  If a "fill-in-able" widget has been found  */

	    widgets_found = 'y';

	/*  If text or radio widget, first value should be chosen if not already  */

	if (select_button == 'y')
	    head->val_list->chosen = TRUE;
    }
    else
        head->next = found_widget(head->next, widt, select_button);

    return (head);
}

/*---------------------------------------------------------------------*/

	/***********************************************************
	*  FUNCTION ERROR_FOUND:                                   *
	*  Build up error message strings and increment the error  *
	*  count.  Takes the length of the current line up to the  *
	*  point of error (int), and a number which represents     *
	*  the type of error (int).  Called from within the Yacc   *
	*  rules whenever an error is found.                       *
	***********************************************************/

error_found(length, error_code)
int  length;
{
    int  i, point_reached;

    error_count++;

    /*  Calculate amount of extra space needed in error mesg. string  */

    point_reached = strlen(error_string);
    if (point_reached > 1)
    {
	error_string[point_reached] = ' ';
	point_reached++;
    }
    error_string = realloc(error_string,(point_reached + length + MARGIN + 2));
    for (i=point_reached;i<(point_reached+length+MARGIN-1);i++)
        error_string[i] = ' ';
    error_string[i] = '\0';
    strcat(error_string,"^\n");

    /*  Calculate amount of extra space needed in error mesg. string  */

    point_reached = strlen(error_string);
    error_string = realloc(error_string,(point_reached + length + MARGIN + 2));
    for (i=point_reached;i<(point_reached+length+MARGIN);i++)
        error_string[i] = ' ';
    error_string[i] = '\0';

    point_reached = strlen(error_string);
    error_string = realloc(error_string, (point_reached + strlen(err_mesgs[error_code])+1));
    strcat(error_string, err_mesgs[error_code]);
}
