.TH XFIELDS 1 "28th April 1989"
.SH NAME
xfields \- generate an X Window System-based form
.SH SYNOPSIS
\fBxfields\fP [\fI-static\fP] [\fI-heading heading\fP] [\fIinputfile\fP [\fIoutputfile\fP]]
.SH DESCRIPTION
Input to \fBxfields\fP is a file containing fields (comments, questions,
variable names and values) from which an X Window System-based form is
generated and displayed on the screen.
The user 'fills in' the form by clicking on buttons or entering text
where appropriate.  The questions need not be answered sequentially
and answers to questions may be revised any number of times before
selecting quit.  Selecting quit causes the user's answers to be
written to the output file, along with the appropriate variable name (see
description of grammar below).  The output is in Bourne shell script format.

The \fI-static\fP flag causes xfields to output the results when the "Okay"
button is pressed but without actually quitting.  This allows the output to
be piped into a Bourne shell "while" construct to continually "eval"uate
the variable settings.
.SH USAGE
The format of the input file is strictly defined.  Each line of the input
file will produce a corresponding entry in the form.  Currently five
types of input line are allowed -  enabling the user to produce
comments (comment lines), questions which require textual answers
(text lines),
questions which require the selection
of zero or more predefined answers (toggle lines),
questions which require the selection of one of a number of 
predefined answers (radio lines), and questions which require a numeric
value in a specified range (value lines).
.LP
The input-file grammar uses \fIstrings\fP and \fIvar_names\fP (see
below), and six reserved symbols: '%', '~', '=', '|', '@' and '#'.
.LP
\fBstrings -\fP
.RS
these may consist of any characters, including tabs
and spaces, but excluding the five reserved symbols.
They are used to construct comment lines, and to form the
question part of any other type of line.
.RE
.LP
\fBvalues -\fP
.RS
same format as strings.  These are used to represent possible answers
to a question.
.RE
.LP
\fBvar_names -\fP
.RS
These must commence with an alphabetic character, followed by zero or
more alphanumeric characters.
They are shell-script variable names, which will hold the answer(s)
selected by users of the form produced by the grammar.
.RE
.sp
.LP
Blank lines and extraneous spaces (ie. those not forming part of
valid strings) in an input file are ignored.  All other lines  must adhere to one
of the formats layed out below.  Note that square brackets enclosing
a symbol mean that that symbol is optional, and curly brackets 
designate zero or more occurences of the pattern they enclose.  '[]'
and '{}' are as such not part of the input-file grammar.
A '#' used before a
value indicates that it is a default answer.  Radio lines
may have only one default answer.  If none is specified, the
first value in the list for that line is made the default answer.
Toggle lines may have any number of default answers.
Text lines may have no default answers,
or one default answer, which will be editable
in the form created.  There is no need to put a '#' before the
default answer in  a text line, although it will do no harm.
Value lines only take three '%' separate arguments. The first two denote
the minimum and maximum range the value can have and the third is the
default value.
.sp
.IP "\fBcomment lines:\fP" 18n
\fIstring\fP
.sp
.IP "\fBtext lines:\fP" 18n
\fIstring~var_name=\fP 
.IP " " 18n
\fIstring~var_name=value\fP
.sp
.IP "\fBvalue lines:\fP" 18n
\fIstring~var_name=min%max%default
.IP "\fBtoggle lines:\fP" 18n
\fIstring~var_name=[#]value|[#]value{|[#]value}
.sp
.IP "\fBradio lines:\fP" 18n
\fIstring~var_name=[#]value@[#]value{@[#]value}
.IP " " 27n
(Zero or one '#' only)
.SH EXAMPLE INPUTFILE
.na
.nj
.nf
Name of Student~name=
College~college=Darwin@Eliot@Keynes@#Rutherford
Term-time address~term_addr=
(Please provide postcode if known)
Funding body~fund=LEA
Length of course~length=3 years@4 years@other
Departments~depts=Hum.|Info. Tec.|Soc. Sci.|Nat. Sci.|Maths.
(Select all departments from which you have followed a course)
Sports played~sports=#tennis|#squash|rugby|football|#weights|other|none
Age~age=14%80%18
.SH AUTHOR
Alison Fowler, University of Kent at Canterbury (amlf@ukc.ac.uk).
With grateful thanks to Richard Hesketh (rlh2@ukc.ac.uk).
April 1988.
.br
-static option and value/number entry add May 1991, Richard Hesketh.
