/* 
 *  XMenu2 - provide simple popup menu for shell scripts... [ xmenu ]
 *
 *  Description:  This "little" toolkit program produces a popup menu on the
 *                screen taking the command line arguments as menu entries.
 *
 *  Features:
 *    OSF/Motif 1.1 support... [ 02-19-91 ]
 *      - new "-menu/-unem" option supports simple & cascading submenus...
 *      - new "-bell" option enables XBell calls...
 *      - new "-quiet" option disables XBell calls...
 *      - new "-punt" option allows setting the default return string...
 *      - new "-geo" option allows setting the popup position of the menu...
 *      - new "-nest" option allows setting the submenu nesting limit...
 *      - new "-xtra" option allows setting the menu/submenu size limit...
 *      - new xtimer mechanism supports the "cancelled" menu condition...
 *
 *      [ OSF/Motif is a trademark of Open Software Foundation, Inc
 *
 *  Copyright 1991 Richard Hess and Consilium Inc.
 *  Copyright 1990 Richard Hesketh / rlh2@ukc.ac.uk
 *                   Computing Lab. University of Kent at Canterbury, UK
 *
 *  See xmenu2.h for the terms of the copyright...
 *
+------------------------------------------------------------------------------
WHO:	Richard Hess			CORP:	Consilium
TITLE:	Staff Engineer			VOICE:	[415] 691-6342
      [ User Interface Services ]	USNAIL:	640 Clyde Court
UUCP:	...!uunet!cimshop!rhess			Mountain View, CA 94043
+------------------------------------------------------------------------------
 */

#include <stdio.h>
#include <Xm/Xm.h>
#include <Xm/RowColumn.h>

#include <Xm/LabelG.h>
#include <Xm/PushBG.h>
#include <Xm/CascadeBG.h>
#include <Xm/SeparatoG.h>

#include "xmenu2.h"

static void buttonPressCbk();
static void unMappedCbk();
static void puntCbk();
static void nestingError();
static void submenuError();
static int  getPrompt();

struct ret {
  String out_val;
  Cardinal exit_status;
};

static Boolean         Bell      = FALSE;
static Display        *dpy;
static XtAppContext    appCon;
static XtIntervalId    xtimer    = NULL;
static char           *DeFault   = NULL;

static XrmOptionDescRec options[] = {
  {"-bell",    ".bell",    XrmoptionNoArg,  (XtPointer)"True"},
  {"-quiet",   ".bell",    XrmoptionNoArg,  (XtPointer)"False"},
  {"-geo",     ".geo",     XrmoptionSepArg, (XtPointer)NULL},
  {"-heading", ".heading", XrmoptionSepArg, (XtPointer)NULL},
  {"-nest",    ".nest",    XrmoptionSepArg, (XtPointer)NULL},
  {"-punt",    ".punt",    XrmoptionSepArg, (XtPointer)NULL},
  {"-xtra",    ".xtra",    XrmoptionSepArg, (XtPointer)NULL},
};

static String          MXdefs[] = {
  "*marginHeight:          1",
  "*marginWidth:           1", 
  "*menuButton.fontList:  -*-helvetica-bold-r-normal--12-120-75-75*iso*",
  "*menuLabel.fontList:   -*-helvetica-medium-r-normal--12-120-75-75*iso*",
  "*menuSub.fontList:     -*-helvetica-medium-r-normal--12-120-75-75*iso*",
  "*menuXtra.fontList:    -*-helvetica-medium-r-normal--12-120-75-75*iso*",
  "*menuXtra.labelString:  Xtra...",
  NULL
  };

/* ****************************************************************[ main ] */

main(argc, argv)
     int argc;
     char *argv[];
{
  Widget        toplevel, shell, button;
  Widget       *menu;
  int          *mcount;
  XrmDatabase   xrdb;
  XtCallbackRec cbk_arg[2];
  XtCallbackRec cbk2_arg[2];
  Arg           args[20];
  int           rtn;
  Cardinal      i, j;
  Position      x, y;
  Window        junk1;
  int           junk2;
  unsigned int  junk3;
  int           flags;
  int           x_rtn, y_rtn;
  unsigned int  w_rtn, h_rtn;
  struct ret   *output_values;
  String        name, Title;
  XrmValue      value;
  char         *blank;
  char         *geo         = NULL;
  int           k           = 0;
  int           iSub        = 0;
  int           nSub        = 0;
  int           nXtra       = 0;
  int           MaxCount;      
  int           NestLimit;      
  
  cbk_arg[0].callback = buttonPressCbk;
  cbk_arg[0].closure = 0;
  cbk_arg[1].callback = NULL;

  cbk2_arg[0].callback = unMappedCbk;
  cbk2_arg[0].closure = 0;
  cbk2_arg[1].callback = NULL;

  XtToolkitInitialize();
  appCon = XtCreateApplicationContext();
  
  XtAppSetFallbackResources(appCon, MXdefs);

  dpy = XtOpenDisplay(appCon, NULL, APP_NAME, APP_CLASS,
		      options, XtNumber(options), &argc, argv);
  
  toplevel = XtAppCreateShell(APP_NAME, APP_CLASS,
			      applicationShellWidgetClass, dpy,
			      NULL, 0);
  xrdb = XtDatabase(dpy);
  
  if (XrmGetResource(xrdb,  XR_BELL, "", &blank, &value)) {
    if (strcmp((char *)value.addr, "True") == 0 )
      Bell = TRUE;
    else
      Bell = FALSE;
  }

  if (XrmGetResource(xrdb,  XR_GEO, "", &blank, &value)) {
    geo = (char *)value.addr;
    flags = XParseGeometry(geo, &x_rtn, &y_rtn,
			   (unsigned int *)&w_rtn,
			   (unsigned int *)&h_rtn);
    if ((XValue & flags) && (YValue & flags)) {
      if (x_rtn < 0)	x = XWidthOfScreen(XtScreen(toplevel)) + x_rtn;
      else		x = x_rtn;
      if (y_rtn < 0)	y = XHeightOfScreen(XtScreen(toplevel)) + y_rtn;
      else		y = y_rtn;
    }
    else 
      x = y = DEFAULT_XY;   /* XValue & YValue search failed...  8-(  */
  }
  else {
    if (XQueryPointer(dpy, DefaultRootWindow(dpy), &junk1, &junk1,
		      &x_rtn, &y_rtn, &junk2, &junk2, &junk3)) {
      x = x_rtn;
      y = y_rtn;
    }
    else {
      x = y = DEFAULT_XY;   /* XQueryPointer() failed...  8-(  */
    }
  }

  if (XrmGetResource(xrdb,  XR_HEADING, "", &blank, &value))
    Title = (char *)value.addr;
  else
    Title = NULL;
  
  if (XrmGetResource(xrdb,  XR_NEST, "", &blank, &value)) {
    NestLimit = atoi((char *)value.addr);
    if (NestLimit < DEFAULT_NEST)
      NestLimit = DEFAULT_NEST;
  }
  else
    NestLimit = DEFAULT_NEST;

  menu = (Widget *) XtMalloc(NestLimit*(sizeof(Widget)));
  mcount = (int *) XtMalloc(NestLimit*(sizeof(int)));

  if (XrmGetResource(xrdb,  XR_PUNT, "", &blank, &value))
    DeFault = (char *)value.addr;
  else
    DeFault = DEFAULT_PUNT;

  if (XrmGetResource(xrdb,  XR_XTRA, "", &blank, &value)) {
    MaxCount = atoi((char *)value.addr);
    if (MaxCount < DEFAULT_MIN)
      MaxCount = DEFAULT_MIN;
  }
  else
    MaxCount = DEFAULT_MAX;

  i = 0;
  XtSetArg(args[i], XmNx, x); i++;
  XtSetArg(args[i], XmNy, y); i++;
  XtSetArg(args[i], XmNunmapCallback, cbk2_arg); i++;
  shell = XmCreatePopupMenu(toplevel, MAIN_MENU, args, i);
  
  menu[nSub] = shell;
  
  if (Title != NULL) {
    i = 0;
    XtSetArg(args[i], XmNlabelString,
	     XmStringCreate(Title, XmSTRING_DEFAULT_CHARSET)); i++;
    button = XmCreateLabelGadget(shell, MENU_LABEL, args, i);
    XtManageChild(button);
    k++;
    button = XmCreateSeparatorGadget(shell, MENU_LINE, NULL, 0);
    XtManageChild(button);
    k++;
  }
  
  for (j = 1; j < argc; j++) {
    rtn = getPrompt(argv[j], &name, &output_values);
    if (k > MaxCount) {
      if (nXtra == 0)
	nXtra = nSub;
      mcount[nSub] = k;
      k = 0;
      nSub++;
      if (nSub == NestLimit)
	nestingError(nSub);
      menu[nSub] = XmCreatePulldownMenu(menu[nSub-1], SUB_MENU, NULL, 0);
      i = 0;
      XtSetArg(args[i], XmNsubMenuId, menu[nSub]); i++;
      button = XmCreateCascadeButtonGadget(menu[nSub-1], MENU_XTRA, args, i);
      XtManageChild(button);
    }
    
    switch (rtn) {
    case LINE:
      button = XmCreateSeparatorGadget(menu[nSub], MENU_LINE, NULL, 0);
      break;
    case MENU:
      mcount[nSub] = k;
      k = 0;
      iSub++;
      nSub++;
      if (nSub == NestLimit)
	nestingError(nSub);
      j++;
      if (j == argc)
	submenuError();
      rtn = getPrompt(argv[j], &name, &output_values);
      switch (rtn) {
      case LINE:
      case MENU:
      case UNEM:
	submenuError();
	break;
      default:
	menu[nSub] = XmCreatePulldownMenu(menu[nSub-1], SUB_MENU, NULL, 0);
	i = 0;
	XtSetArg(args[i], XmNsubMenuId, menu[nSub]); i++;
	XtSetArg(args[i], XmNlabelString,
		 XmStringCreate(name, XmSTRING_DEFAULT_CHARSET)); i++;
	button = XmCreateCascadeButtonGadget(menu[nSub-1], MENU_SUB, args, i);
	break;
      }
      break;
    case UNEM:
      iSub--;
      if (nXtra && iSub < nXtra) {
	nSub = nXtra-1;
	nXtra = 0;
      }
      else
	nSub--;
      if (iSub < 0) {
	if (Bell) {
	  XBell(dpy, 50);
	  XFlush(dpy);
	}
	fprintf(stderr,
		"xmMenu:  unbalanced submenu closure... [ %s ]\n", UNEM_TOK);
	exit(1);
      }
      k = mcount[nSub];
      break;
    default:
      i = 0;
      XtSetArg(args[i], XmNlabelString,
	       XmStringCreate(name, XmSTRING_DEFAULT_CHARSET)); i++;
      cbk_arg[0].closure = (XtPointer)output_values;
      XtSetArg(args[i], XmNactivateCallback, cbk_arg); i++;
      button = XmCreatePushButtonGadget(menu[nSub], MENU_BUTTON, args, i);
      break;
    }
    XtManageChild(button);
    k++;
  }
  if (!k) {
    if (Bell) {
      XBell(dpy, 50);
      XFlush(dpy);
    }
    fprintf(stderr, "xmMenu:  no menu items given...\n");
    exit(1);
  }
  
  XtManageChild(shell);
  
  XtAppMainLoop(appCon);
}

/* ******************************************************[ buttonPressCbk ] */

static void buttonPressCbk(w, client_data, call_data)
     Widget w;
     XtPointer client_data, call_data;
{
  struct ret *values;
  
  if (xtimer) {
    XtRemoveTimeOut(xtimer);
    xtimer = NULL;
  }

  values = (struct ret *)client_data;
  if (values->out_val != NULL)
    fprintf(stdout, "%s\n", values->out_val);
  exit(values->exit_status);
}

/* *********************************************************[ unMappedCbk ] */

static void unMappedCbk(w, client_data, call_data)
     Widget w;
     XtPointer client_data, call_data;
{
  xtimer = XtAppAddTimeOut(appCon, XTIMEOUT, puntCbk, 0);
}

/* *************************************************************[ puntCbk ] */

static void puntCbk(data, id)
     XtPointer data;
     XtIntervalId *id;
{
  if (Bell) {
    XBell(dpy, 50);
    XFlush(dpy);
  }
  if (DeFault != NULL)
    fprintf(stdout, "%s\n", DeFault);
  exit(0);
}

/* ********************************************************[ nestingError ] */

static void nestingError(nest)
     int nest;
{
  if (Bell) {
    XBell(dpy, 50);
    XFlush(dpy);
  }
  fprintf(stderr,
	  "xmMenu:  submenu nesting limit exceeded... [ %d ]\n",
	  nest);
  exit(1);
}

/* ********************************************************[ submenuError ] */

static void submenuError()
{
  if (Bell) {
    XBell(dpy, 50);
    XFlush(dpy);
  }
  fprintf(stderr, "xmMenu:  invalid submenu definition... [ no label ]\n");
  exit(1);
}

/* ***********************************************************[ getPrompt ] */

static int getPrompt(str, rtn_name, values)
     String str;
     String *rtn_name;
     struct ret **values;
{
  static Cardinal count = STATUS_START;
  int s, t;
  char name[1000];
  Boolean found_escape = FALSE;
  
  *values = NULL;
  
  if (strncmp(str, LINE_TOK, strlen(LINE_TOK)) == 0) {
    rtn_name = NULL;
    return (LINE);
  }
  if (strncmp(str, MENU_TOK, strlen(MENU_TOK)) == 0) {
    rtn_name = NULL;
    return (MENU);
  }
  if (strncmp(str, UNEM_TOK, strlen(UNEM_TOK)) == 0) {
    rtn_name = NULL;
    return (UNEM);
  }
  *values = XtNew(struct ret);
  (*values)->out_val = NULL;
  
  for (t = s = 0; str[s] != '\0'; s++) {
    if (str[s] == '\\') {
      if (found_escape) {
	found_escape = FALSE;
	name[t++] = '\\';
      } else
	found_escape = TRUE;
    } else if (str[s] == '=') {
      if (found_escape) {
	found_escape = FALSE;
	name[t++] = '=';
      } else {
	(*values)->out_val = str + s + 1;
	break;
      }
    } else
      name[t++] = str[s];
  }
  name[t] = '\0';
  if ((*values)->out_val == NULL)
    (*values)->out_val = str;
  
  (*values)->exit_status = count++;
  *rtn_name = XtNewString(name);
  return (0);
}

/* *******************************************************************<eof> */
